"""Refresh the authorization token for new credentials."""

import logging
import traceback
from urllib.parse import parse_qs # noqa pylint: disable=no-name-in-module, import-error

from shared import (create_error_html, extract_and_parse_cookies, get_config,  # noqa pylint: disable=import-error
                    get_cookie_headers, http_post_with_retry)

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()


def handler(event, _context):
    """Handle the authorization refresh.

    Keyword Args:
        event: The Lambda Event
    """
    request = event['Records'][0]['cf']['request']
    domain_name = request['headers']['host'][0]['value']
    redirected_from_uri = 'https://%s' % domain_name

    try:
        parsed_qs = parse_qs(request.get('querystring'))
        requested_uri = parsed_qs.get('requestedUri')[0]
        current_nonce = parsed_qs.get('nonce')[0]
        # Add the requested uri path to the main
        redirected_from_uri += requested_uri or ""

        cookies = extract_and_parse_cookies(
            request.get('headers'),
            CONFIG.get('client_id')
        )

        tokens = {
            'id_token': cookies.get('idToken'),
            'access_token': cookies.get('accessToken'),
            'refresh_token': cookies.get('refreshToken')
        }

        validate_refresh_request(
            current_nonce,
            cookies.get('nonce'),
            tokens
        )

        try:
            # Request new tokens based on the refresh_token
            body = {
                'grant_type': 'refresh_token',
                'client_id': CONFIG.get('client_id'),
                'refresh_token': tokens.get('refresh_token')
            }
            res = http_post_with_retry(
                ('https://%s/oauth2/token' % CONFIG.get('cognito_auth_domain')),
                body,
                {'Content-Type': 'application/x-www-form-urlencoded'}
            )
            tokens['id_token'] = res.get('id_token')
            tokens['access_token'] = res.get('access_token')
        except Exception as err:  # pylint: disable=broad-except
            LOGGER.error(err)
            # Otherwise clear the refresh token
            tokens['refresh_token'] = ""

        headers = {
            'location': [
                {
                    'key': 'location',
                    'value': redirected_from_uri
                }
            ],
            'set-cookie': get_cookie_headers(
                CONFIG.get('client_id'),
                CONFIG.get('oauth_scopes'),
                tokens,
                domain_name,
                CONFIG.get('cookie_settings')
            )
        }
        headers.update(CONFIG.get('cloud_front_headers'))

        # Redirect the user back to their requested uri
        # with new tokens at hand
        return {
            'status': '307',
            'statusDescription': 'Temporary Redirect',
            'headers': headers
        }

    # Send a basic html error response and inform the user
    # why refresh was unsuccessful
    except Exception as err:  # pylint: disable=broad-except
        LOGGER.info(err)
        LOGGER.info(traceback.print_exc())

        headers = {
            "content-type": [
                {
                    "key": "Content-Type",
                    "value": "text/html; charset=UTF-8"
                }
            ]
        }
        headers.update(CONFIG.get('cloud_front_headers'))

        return {
            'body': create_error_html("Bad Request", err, redirected_from_uri),
            'status': '400',
            'headers': headers
        }


def validate_refresh_request(current_nonce,
                             original_nonce,
                             tokens):
    """Validate that nonce and tokens are present.

    Keyword Args:
        current_nonce (str): The current nonce code
        original_nonce (str): The original nonce code
        tokens (Dict[str, str]): A dictionary of all the token_types
            and their corresponding token values (id, auth, refresh)
    """
    if not original_nonce:
        msg = "Your browser didn't send the nonce cookie along, " + \
            "but it is required for security (prevent CSRF)."
        LOGGER.error(msg)
        raise Exception(msg)

    if current_nonce != original_nonce:
        msg = "Nonce mismatch"
        LOGGER.error(msg)
        raise Exception(msg)

    for token_type, token in tokens.items():
        if not token:
            msg = "Missing %s" % token_type
            LOGGER.error(msg)
            raise Exception(msg)



"""Tests for runway.cfngin.plan."""
# pylint: disable=protected-access,unused-argument
import json
import os
import shutil
import tempfile
import unittest

import mock

from runway.cfngin.context import Config, Context
from runway.cfngin.dag import walk
from runway.cfngin.exceptions import (CancelExecution, GraphError,
                                      PersistentGraphLocked, PlanFailed)
from runway.cfngin.lookups.registry import (register_lookup_handler,
                                            unregister_lookup_handler)
from runway.cfngin.plan import Graph, Plan, Step
from runway.cfngin.stack import Stack
from runway.cfngin.status import COMPLETE, FAILED, SKIPPED, SUBMITTED
from runway.cfngin.util import stack_template_key_name

from .factories import generate_definition, mock_context


class TestStep(unittest.TestCase):
    """Tests for runway.cfngin.plan.Step."""

    def setUp(self):
        """Run before tests."""
        stack = mock.MagicMock()
        stack.name = "stack"
        stack.fqn = "namespace-stack"
        self.step = Step(stack=stack, fn=None)

    def test_status(self):
        """Test status."""
        self.assertFalse(self.step.submitted)
        self.assertFalse(self.step.completed)

        self.step.submit()
        self.assertEqual(self.step.status, SUBMITTED)
        self.assertTrue(self.step.submitted)
        self.assertFalse(self.step.completed)

        self.step.complete()
        self.assertEqual(self.step.status, COMPLETE)
        self.assertNotEqual(self.step.status, SUBMITTED)
        self.assertTrue(self.step.submitted)
        self.assertTrue(self.step.completed)

        self.assertNotEqual(self.step.status, True)
        self.assertNotEqual(self.step.status, False)
        self.assertNotEqual(self.step.status, 'banana')

    def test_from_stack_name(self):
        """Return step from step name."""
        context = mock_context()
        stack_name = 'test-stack'
        result = Step.from_stack_name(stack_name, context)

        self.assertIsInstance(result, Step)
        self.assertEqual(stack_name, result.stack.name)

    def test_from_persistent_graph(self):
        """Return list of steps from graph dict."""
        context = mock_context()
        graph_dict = {
            'stack1': [],
            'stack2': ['stack1']
        }
        result = Step.from_persistent_graph(graph_dict, context)

        self.assertEqual(2, len(result))
        self.assertIsInstance(result, list)

        for step in result:
            self.assertIsInstance(step, Step)
            self.assertIn(step.stack.name, graph_dict.keys())


class TestGraph(unittest.TestCase):
    """Tests for runway.cfngin.plan.Graph."""

    def setUp(self):
        """Run before tests."""
        self.context = mock_context()
        self.graph_dict = {
            'stack1': [],
            'stack2': ['stack1']
        }
        self.graph_dict_expected = {
            'stack1': set(),
            'stack2': set(['stack1'])
        }
        self.steps = Step.from_persistent_graph(self.graph_dict,
                                                self.context)

    def test_add_steps(self):
        """Test add steps."""
        graph = Graph()
        graph.add_steps(self.steps)

        self.assertEqual(self.steps, list(graph.steps.values()))
        self.assertEqual([step.name for step in self.steps],
                         list(graph.steps.keys()))
        self.assertEqual(self.graph_dict_expected, graph.to_dict())

    def test_pop(self):
        """Test pop."""
        graph = Graph()
        graph.add_steps(self.steps)

        stack2 = next(step for step in self.steps if step.name == 'stack2')

        self.assertEqual(stack2, graph.pop(stack2))
        self.assertEqual({'stack1': set()}, graph.to_dict())

    def test_dumps(self):
        """Test dumps."""
        graph = Graph()
        graph.add_steps(self.steps)

        self.assertEqual(json.dumps(self.graph_dict), graph.dumps())

    def test_from_dict(self):
        """Test from dict."""
        graph = Graph.from_dict(self.graph_dict, self.context)

        self.assertIsInstance(graph, Graph)
        self.assertEqual([step.name for step in self.steps],
                         list(graph.steps.keys()))
        self.assertEqual(self.graph_dict_expected, graph.to_dict())

    def test_from_steps(self):
        """Test from steps."""
        graph = Graph.from_steps(self.steps)

        self.assertEqual(self.steps, list(graph.steps.values()))
        self.assertEqual([step.name for step in self.steps],
                         list(graph.steps.keys()))
        self.assertEqual(self.graph_dict_expected, graph.to_dict())


class TestPlan(unittest.TestCase):
    """Tests for runway.cfngin.plan.Plan."""

    def setUp(self):
        """Run before tests."""
        self.count = 0
        self.config = Config({"namespace": "namespace"})
        self.context = Context(config=self.config)
        register_lookup_handler("noop", lambda **kwargs: "test")

    def tearDown(self):
        """Run after tests."""
        unregister_lookup_handler("noop")

    def test_plan(self):
        """Test plan."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=self.context)

        graph = Graph.from_steps([Step(vpc, fn=None), Step(bastion, fn=None)])
        plan = Plan(description="Test", graph=graph)

        self.assertEqual(plan.graph.to_dict(), {
            'bastion.1': set(['vpc.1']),
            'vpc.1': set([])})

    def test_plan_reverse(self):
        """Test plan reverse."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=self.context)
        graph = Graph.from_steps([Step(vpc, fn=None), Step(bastion, fn=None)])
        plan = Plan(description="Test", graph=graph, reverse=True)

        # order is different between python2/3 so can't compare dicts
        result_graph_dict = plan.graph.to_dict()
        self.assertEqual(set(), result_graph_dict.get('bastion.1'))
        self.assertEqual(set(['bastion.1']), result_graph_dict.get('vpc.1'))

    def test_plan_targeted(self):
        """Test plan targeted."""
        context = Context(config=self.config)
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=context)
        context.stack_names = [vpc.name]
        graph = Graph.from_steps([Step(vpc, fn=None), Step(bastion, fn=None)])
        plan = Plan(description="Test", graph=graph, context=context)

        self.assertEqual({vpc.name: set()}, plan.graph.to_dict())

    def test_execute_plan(self):
        """Test execute plan."""
        context = Context(config=self.config)
        context.put_persistent_graph = mock.MagicMock()
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=context)
        removed = Stack(
            definition=generate_definition('removed', 1, requires=[]),
            context=context)
        context._persistent_graph = Graph.from_steps([removed])

        calls = []

        def _launch_stack(stack, status=None):
            calls.append(stack.fqn)
            return COMPLETE

        def _destroy_stack(stack, status=None):
            calls.append(stack.fqn)
            return COMPLETE

        graph = Graph.from_steps([Step(removed, _destroy_stack),
                                  Step(vpc, _launch_stack),
                                  Step(bastion, _launch_stack)])
        plan = Plan(description="Test", graph=graph, context=context)
        plan.context._persistent_graph_lock_code = plan.lock_code
        plan.execute(walk)

        # the order these are appended changes between python2/3
        self.assertIn('namespace-vpc.1', calls)
        self.assertIn('namespace-bastion.1', calls)
        self.assertIn('namespace-removed.1', calls)
        context.put_persistent_graph.assert_called()

        # order is different between python2/3 so can't compare dicts
        result_graph_dict = context.persistent_graph.to_dict()
        self.assertEqual(2, len(result_graph_dict))
        self.assertEqual(set(), result_graph_dict.get('vpc.1'))
        self.assertEqual(set(['vpc.1']), result_graph_dict.get('bastion.1'))
        self.assertIsNone(result_graph_dict.get('namespace-removed.1'))

    def test_execute_plan_no_persist(self):
        """Test execute plan with no persistent graph."""
        context = Context(config=self.config)
        context.put_persistent_graph = mock.MagicMock()
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=context)

        calls = []

        def _launch_stack(stack, status=None):
            calls.append(stack.fqn)
            return COMPLETE

        graph = Graph.from_steps([Step(vpc, _launch_stack),
                                  Step(bastion, _launch_stack)])
        plan = Plan(description="Test", graph=graph, context=context)

        plan.execute(walk)

        self.assertEqual(calls, ['namespace-vpc.1', 'namespace-bastion.1'])
        context.put_persistent_graph.assert_not_called()

    def test_execute_plan_locked(self):
        """Test execute plan locked.

        Locked stacks still need to have their requires evaluated when
        they're being created.

        """
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            locked=True,
            context=self.context)

        calls = []

        def fn(stack, status=None):
            calls.append(stack.fqn)
            return COMPLETE

        graph = Graph.from_steps([Step(vpc, fn), Step(bastion, fn)])
        plan = Plan(description="Test", graph=graph)
        plan.execute(walk)

        self.assertEqual(calls, ['namespace-vpc.1', 'namespace-bastion.1'])

    def test_execute_plan_filtered(self):
        """Test execute plan filtered."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        db = Stack(
            definition=generate_definition('db', 1, requires=[vpc.name]),
            context=self.context)
        app = Stack(
            definition=generate_definition('app', 1, requires=[db.name]),
            context=self.context)

        calls = []

        def fn(stack, status=None):
            calls.append(stack.fqn)
            return COMPLETE

        context = mock.MagicMock()
        context.persistent_graph_locked = False
        context.stack_names = ['db.1']
        graph = Graph.from_steps([Step(vpc, fn), Step(db, fn), Step(app, fn)])
        plan = Plan(
            context=context,
            description="Test",
            graph=graph)
        plan.execute(walk)

        self.assertEqual(calls, [
            'namespace-vpc.1', 'namespace-db.1'])

    def test_execute_plan_exception(self):
        """Test execute plan exception."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=self.context)

        calls = []

        def fn(stack, status=None):
            calls.append(stack.fqn)
            if stack.name == vpc_step.name:
                raise ValueError('Boom')
            return COMPLETE

        vpc_step = Step(vpc, fn)
        bastion_step = Step(bastion, fn)

        graph = Graph.from_steps([vpc_step, bastion_step])
        plan = Plan(description="Test", graph=graph)

        with self.assertRaises(PlanFailed):
            plan.execute(walk)

        self.assertEqual(calls, ['namespace-vpc.1'])
        self.assertEqual(vpc_step.status, FAILED)

    def test_execute_plan_skipped(self):
        """Test execute plan skipped."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=self.context)

        calls = []

        def fn(stack, status=None):
            calls.append(stack.fqn)
            if stack.fqn == vpc_step.name:
                return SKIPPED
            return COMPLETE

        vpc_step = Step(vpc, fn)
        bastion_step = Step(bastion, fn)

        graph = Graph.from_steps([vpc_step, bastion_step])
        plan = Plan(description="Test", graph=graph)
        plan.execute(walk)

        self.assertEqual(calls, ['namespace-vpc.1', 'namespace-bastion.1'])

    def test_execute_plan_failed(self):
        """Test execute plan failed."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=self.context)
        db = Stack(
            definition=generate_definition('db', 1),
            context=self.context)

        calls = []

        def fn(stack, status=None):
            calls.append(stack.fqn)
            if stack.name == vpc_step.name:
                return FAILED
            return COMPLETE

        vpc_step = Step(vpc, fn)
        bastion_step = Step(bastion, fn)
        db_step = Step(db, fn)

        graph = Graph.from_steps([vpc_step, bastion_step, db_step])
        plan = Plan(description="Test", graph=graph)
        with self.assertRaises(PlanFailed):
            plan.execute(walk)

        calls.sort()

        self.assertEqual(calls, ['namespace-db.1', 'namespace-vpc.1'])

    def test_execute_plan_cancelled(self):
        """Test execute plan cancelled."""
        vpc = Stack(
            definition=generate_definition('vpc', 1),
            context=self.context)
        bastion = Stack(
            definition=generate_definition('bastion', 1, requires=[vpc.name]),
            context=self.context)

        calls = []

        def fn(stack, status=None):
            calls.append(stack.fqn)
            if stack.fqn == vpc_step.name:
                raise CancelExecution
            return COMPLETE

        vpc_step = Step(vpc, fn)
        bastion_step = Step(bastion, fn)

        graph = Graph.from_steps([vpc_step, bastion_step])
        plan = Plan(description="Test", graph=graph)
        plan.execute(walk)

        self.assertEqual(calls, ['namespace-vpc.1', 'namespace-bastion.1'])

    def test_execute_plan_graph_locked(self):
        """Test execute plan with locked persistent graph."""
        context = Context(config=self.config)
        context._persistent_graph = Graph.from_dict({'stack1': []}, context)
        context._persistent_graph_lock_code = '1111'
        plan = Plan(description='Test', graph=Graph(), context=context)
        print(plan.locked)

        with self.assertRaises(PersistentGraphLocked):
            plan.execute()

    def test_build_graph_missing_dependency(self):
        """Test build graph missing dependency."""
        bastion = Stack(
            definition=generate_definition(
                'bastion', 1, requires=['vpc.1']),
            context=self.context)

        with self.assertRaises(GraphError) as expected:
            Graph.from_steps([Step(bastion, None)])
        message_starts = (
            "Error detected when adding 'vpc.1' "
            "as a dependency of 'bastion.1':"
        )
        message_contains = "dependent node vpc.1 does not exist"
        self.assertTrue(str(expected.exception).startswith(message_starts))
        self.assertTrue(message_contains in str(expected.exception))

    def test_build_graph_cyclic_dependencies(self):
        """Test build graph cyclic dependencies."""
        vpc = Stack(
            definition=generate_definition(
                'vpc', 1),
            context=self.context)
        db = Stack(
            definition=generate_definition(
                'db', 1, requires=['app.1']),
            context=self.context)
        app = Stack(
            definition=generate_definition(
                'app', 1, requires=['db.1']),
            context=self.context)

        with self.assertRaises(GraphError) as expected:
            Graph.from_steps([Step(vpc, None), Step(db, None), Step(app, None)])
        message = ("Error detected when adding 'db.1' "
                   "as a dependency of 'app.1': graph is "
                   "not acyclic")
        self.assertEqual(str(expected.exception), message)

    def test_dump(self, *args):
        """Test dump."""
        requires = None
        steps = []

        for i in range(5):
            overrides = {
                "variables": {
                    "PublicSubnets": "1",
                    "SshKeyName": "1",
                    "PrivateSubnets": "1",
                    "Random": "${noop something}",
                },
                "requires": requires,
            }

            stack = Stack(
                definition=generate_definition('vpc', i, **overrides),
                context=self.context)
            requires = [stack.name]

            steps += [Step(stack, None)]

        graph = Graph.from_steps(steps)
        plan = Plan(description="Test", graph=graph)

        tmp_dir = tempfile.mkdtemp()
        try:
            plan.dump(tmp_dir, context=self.context)

            for step in plan.steps:
                template_path = os.path.join(
                    tmp_dir,
                    stack_template_key_name(step.stack.blueprint))
                self.assertTrue(os.path.isfile(template_path))
        finally:
            shutil.rmtree(tmp_dir)



"""Test runway.lookups.handlers.ssm."""
# pylint: disable=no-self-use,unused-import
import json
from datetime import datetime

import pytest
import yaml

from runway.cfngin.exceptions import FailedVariableLookup
from runway.variables import Variable


def get_parameter_response(name, value, value_type='String', label=None,
                           version=1):
    """Generate a mock ssm.get_parameter response."""
    selector = '{}/{}'.format(name, label or version)
    return {
        'Parameter': {
            'Name': name,
            'Type': value_type,
            'Value': value,
            'Version': 1,
            'Selector': selector,
            'SourceResult': '',
            'LastModifiedDate': datetime.now(),
            'ARN': ''
        }
    }


def get_parameter_request(name, decrypt=True):
    """Generate the expected request paramters for ssm.get_parameter."""
    return {
        'Name': name,
        'WithDecryption': decrypt
    }


class TestSsmLookup(object):
    """Test runway.lookups.handlers.ssm.SsmLookup."""

    def test_basic(self, cfngin_context, runway_context):
        """Test resolution of a basic lookup."""
        name = '/test/param'
        value = 'test value'
        cfngin_stubber = cfngin_context.add_stubber('ssm')
        runway_stubber = runway_context.add_stubber('ssm')
        cfngin_var = Variable('test_var', '${ssm %s}' % name,
                              variable_type='cfngin')
        runway_var = Variable('test_var', '${ssm %s}' % name,
                              variable_type='runway')

        for stubber in [cfngin_stubber, runway_stubber]:
            stubber.add_response('get_parameter',
                                 get_parameter_response(name, value),
                                 get_parameter_request(name))

        with cfngin_stubber as cfn_stub, runway_stubber as rw_stub:
            cfngin_var.resolve(context=cfngin_context)
            assert cfngin_var.value == value

            runway_var.resolve(context=runway_context)
            assert runway_var.value == value

        cfn_stub.assert_no_pending_responses()
        rw_stub.assert_no_pending_responses()

    def test_default(self, runway_context):
        """Test resolution of a default value."""
        name = '/test/param'
        value = 'test value'
        stubber = runway_context.add_stubber('ssm')
        var = Variable(
            'test_var',
            '${ssm /test/invalid::load=json, default=${ssm %s}}' % name,
            variable_type='runway'
        )

        stubber.add_response('get_parameter',
                             get_parameter_response(name, value),
                             get_parameter_request(name))
        stubber.add_client_error('get_parameter',
                                 'ParameterNotFound',
                                 expected_params=get_parameter_request(
                                     '/test/invalid'
                                 ))

        with stubber as stub:
            var.resolve(context=runway_context)
            assert var.value == value
            stub.assert_no_pending_responses()

    def test_different_region(self, runway_context):
        """Test Lookup in region other than that set in Context."""
        name = '/test/param'
        value = 'test value'
        stubber = runway_context.add_stubber('ssm', region='us-west-2')
        var = Variable('test_var', '${ssm %s::region=us-west-2}' % name,
                       variable_type='runway')

        stubber.add_response('get_parameter',
                             get_parameter_response(name, value),
                             get_parameter_request(name))

        with stubber as stub:
            var.resolve(context=runway_context)
            assert var.value == value
            stub.assert_no_pending_responses()

    def test_loaded_value(self, runway_context):
        """Test resolution of a JSON value."""
        name = '/test/param'
        raw_value = {
            'nested': {
                'bool': True,
                'nest_key': 'nested_val'
            },
            'test_key': 'test_val'
        }
        stubber = runway_context.add_stubber('ssm')
        parsers = ['json', 'yaml']
        tests = [
            {
                'lookup': '${{ssm {name}::load={parser}}}',
                'expected': raw_value
            },
            {
                'lookup': '${{ssm {name}::load={parser},transform=str,indent=2}}',
                'expected': json.dumps(json.dumps(raw_value, indent=2))
            },
            {
                'lookup': '${{ssm {name}::load={parser},get=nested}}',
                'expected': raw_value['nested']
            },
            {
                'lookup': '${{ssm {name}::load={parser},get=nested.bool,transform=str}}',
                'expected': json.dumps('True')
            }
        ]

        for parser in parsers:
            for test in tests:
                var = Variable('test_var.{}'.format(parser),
                               test['lookup'].format(name=name, parser=parser),
                               variable_type='runway')
                if parser == 'json':
                    dumped_value = json.dumps(raw_value)
                elif parser == 'yaml':
                    dumped_value = yaml.dump(raw_value)
                else:
                    raise ValueError

                stubber.add_response('get_parameter',
                                     get_parameter_response(name,
                                                            dumped_value),
                                     get_parameter_request(name))

                with stubber as stub:
                    var.resolve(context=runway_context)
                    assert var.value == test['expected']
                    stub.assert_no_pending_responses()

    def test_not_found(self, runway_context):
        """Test raises ParameterNotFound."""
        name = '/test/param'
        stubber = runway_context.add_stubber('ssm')
        var = Variable('test_var', '${ssm %s}' % name, variable_type='runway')

        stubber.add_client_error('get_parameter',
                                 'ParameterNotFound',
                                 expected_params=get_parameter_request(
                                     name
                                 ))

        with stubber as stub, pytest.raises(FailedVariableLookup) as err:
            var.resolve(context=runway_context)

        assert 'ParameterNotFound' in str(err.value)
        stub.assert_no_pending_responses()



"""Tests for runway.cfngin.hooks.command."""
import os
import unittest
from subprocess import PIPE

import mock

from runway.cfngin.config import Config
from runway.cfngin.context import Context
from runway.cfngin.hooks.command import run_command

from ..factories import mock_provider


class MockProcess(object):
    """Mock process."""

    def __init__(self, returncode=0, stdout='', stderr=''):
        """Instantiate class."""
        self.returncode = returncode
        self.stdout = stdout
        self.stderr = stderr
        self.stdin = None

    def communicate(self, stdin):
        """Communicate with process."""
        self.stdin = stdin
        return (self.stdout, self.stderr)

    def wait(self):
        """Wait for process."""
        return self.returncode

    def kill(self):
        """Kill process."""
        return


class TestCommandHook(unittest.TestCase):
    """Tests for runway.cfngin.hooks.command."""

    def setUp(self):
        """Run before tests."""
        self.context = Context(
            config=Config({'namespace': 'test', 'stacker_bucket': 'test'}))
        self.provider = mock_provider(region="us-east-1")

        self.mock_process = MockProcess()
        self.popen_mock = \
            mock.patch('runway.cfngin.hooks.command.Popen',
                       return_value=self.mock_process).start()

        self.devnull = mock.Mock()
        self.devnull_mock = \
            mock.patch('runway.cfngin.hooks.command._devnull',
                       return_value=self.devnull).start()

    def tearDown(self):
        """Run after tests."""
        self.devnull_mock.stop()
        self.popen_mock.stop()

    def run_hook(self, **kwargs):
        """Run hook."""
        real_kwargs = {
            'context': self.context,
            'provider': self.provider,
        }
        real_kwargs.update(kwargs)

        return run_command(**real_kwargs)

    def test_command_ok(self):
        """Test command ok."""
        self.mock_process.returncode = 0
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        results = self.run_hook(command=['foo'])

        self.assertEqual(
            results, {'returncode': 0, 'stdout': None, 'stderr': None})
        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=self.devnull, stdout=None, stderr=None, env=None)

    def test_command_fail(self):
        """Test command fail."""
        self.mock_process.returncode = 1
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        results = self.run_hook(command=['foo'])

        self.assertEqual(results, None)
        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=self.devnull, stdout=None, stderr=None, env=None)

    def test_command_ignore_status(self):
        """Test command ignore status."""
        self.mock_process.returncode = 1
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        results = self.run_hook(command=['foo'], ignore_status=True)

        self.assertEqual(
            results, {'returncode': 1, 'stdout': None, 'stderr': None})
        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=self.devnull, stdout=None, stderr=None, env=None)

    def test_command_quiet(self):
        """Test command quiet."""
        self.mock_process.returncode = 0
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        results = self.run_hook(command=['foo'], quiet=True)
        self.assertEqual(
            results, {'returncode': 0, 'stdout': None, 'stderr': None})

        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=self.devnull, stdout=self.devnull,
            stderr=self.devnull, env=None)

    def test_command_interactive(self):
        """Test command interactive."""
        self.mock_process.returncode = 0
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        results = self.run_hook(command=['foo'], interactive=True)
        self.assertEqual(
            results, {'returncode': 0, 'stdout': None, 'stderr': None})

        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=None, stdout=None, stderr=None, env=None)

    def test_command_input(self):
        """Test command input."""
        self.mock_process.returncode = 0
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        results = self.run_hook(command=['foo'], stdin='hello world')
        self.assertEqual(
            results, {'returncode': 0, 'stdout': None, 'stderr': None})

        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=PIPE, stdout=None, stderr=None, env=None)
        self.assertEqual(self.mock_process.stdin, 'hello world')

    def test_command_capture(self):
        """Test command capture."""
        self.mock_process.returncode = 0
        self.mock_process.stdout = 'hello'
        self.mock_process.stderr = 'world'

        results = self.run_hook(command=['foo'], capture=True)
        self.assertEqual(
            results, {'returncode': 0, 'stdout': 'hello', 'stderr': 'world'})

        self.popen_mock.assert_called_once_with(
            ['foo'], stdin=self.devnull, stdout=PIPE, stderr=PIPE, env=None)

    def test_command_env(self):
        """Test command env."""
        self.mock_process.returncode = 0
        self.mock_process.stdout = None
        self.mock_process.stderr = None

        with mock.patch.dict(os.environ, {'foo': 'bar'}, clear=True):
            results = self.run_hook(command=['foo'], env={'hello': 'world'})

            self.assertEqual(results, {'returncode': 0,
                                       'stdout': None,
                                       'stderr': None})
            self.popen_mock.assert_called_once_with(
                ['foo'], stdin=self.devnull, stdout=None, stderr=None,
                env={'hello': 'world', 'foo': 'bar'})



"""Tests for lookup handler base class."""
# pytest: disable=no-self-use
import json
from unittest import TestCase

import yaml

from runway.lookups.handlers.base import LookupHandler
from runway.util import MutableMap


class TestLookupHandler(TestCase):
    """Tests for LookupHandler."""

    def test_abstract_handle(self):
        """Handle should not be implimented."""
        with self.assertRaises(NotImplementedError):
            LookupHandler.handle(None, None)

    def test_dependencies(self):
        """Test dependencies.

        This should always return an empty set.

        """
        assert isinstance(LookupHandler.dependencies({}), set)

    def test_format_results(self):
        """Test format_results."""
        test_dict = {
            'nested': {
                'bool': True,
                'nested_key': 'nested_value'
            },
            'test_key': 'test_value'
        }
        mute_map = MutableMap(**test_dict.copy())

        assert LookupHandler.format_results(test_dict) == test_dict
        assert LookupHandler.format_results(mute_map) == test_dict

        assert LookupHandler.format_results(
            test_dict, get='test_key'
        ) == test_dict['test_key']
        assert LookupHandler.format_results(
            mute_map, get='test_key'
        ) == mute_map.test_key
        assert LookupHandler.format_results(
            mute_map, get='nested'
        ) == mute_map.nested.data
        assert LookupHandler.format_results(
            mute_map, get='nested.nested_key'
        ) == mute_map.nested.nested_key
        assert LookupHandler.format_results(mute_map, get='nested.bool')

        assert LookupHandler.format_results(
            mute_map, transform='str'
        ) == json.dumps(json.dumps(test_dict, indent=0))
        assert LookupHandler.format_results(
            mute_map, transform='str', indent=2
        ) == json.dumps(json.dumps(test_dict, indent=2))
        assert LookupHandler.format_results(
            mute_map, get='nested.bool', transform='str'
        ) == '"True"'

        with self.assertRaises(TypeError):
            LookupHandler.format_results(['something'], get='key')

    def test_load_no_parser(self):
        """Test load with no parser."""
        assert LookupHandler.load('something') == 'something'

    def test_load_list(self):
        """Test load parsers with a list result."""
        value = ['something', 'something-else']
        assert LookupHandler.load(json.dumps(value), parser='json') == value
        assert LookupHandler.load(yaml.dump(value), parser='yaml') == value

    def test_parse(self):
        """Basic value parsing."""
        expected_query = 'my_query'

        result_query, result_args = LookupHandler.parse(expected_query)

        self.assertEqual(result_query, expected_query)
        self.assertEqual(result_args, {})

    def test_parse_args(self):
        """Parse query and args from value."""
        expected_args = {
            'key1': 'val1'
        }
        expected_query = 'my_query'
        value = '{}::{}'.format(expected_query, ','.join([
            '{}={}'.format(key, val) for key, val in expected_args.items()
        ]))

        result_query, result_args = LookupHandler.parse(value)

        self.assertEqual(result_query, expected_query)
        self.assertEqual(result_args, expected_args)

    def test_transform_bool_to_bool(self):
        """Bool should be returned as is."""
        result_true = LookupHandler.transform(True, to_type='bool')
        result_false = LookupHandler.transform(False, to_type='bool')

        self.assertTrue(result_true)
        self.assertFalse(result_false)

    def test_transform_no_type(self):
        """Test transform with no type."""
        assert isinstance(LookupHandler.transform(True, to_type=None), bool)
        assert isinstance(LookupHandler.transform(['something'],
                                                  to_type=None), list)
        assert isinstance(LookupHandler.transform('something',
                                                  to_type=None), str)

    def test_transform_str_to_bool(self):
        """String should be transformed using strtobool."""
        result_true = LookupHandler.transform('true', to_type='bool')
        result_false = LookupHandler.transform('false', to_type='bool')

        self.assertTrue(result_true)
        self.assertFalse(result_false)

    def test_transform_type_check(self):
        """Transform to bool type check."""
        with self.assertRaises(TypeError, msg='dict should raise an error'):
            LookupHandler.transform({'key1': 'val1'}, to_type='bool')

        with self.assertRaises(TypeError, msg='list should raise an error'):
            LookupHandler.transform(['li1'], to_type='bool')

        with self.assertRaises(TypeError, msg='number should raise an error'):
            LookupHandler.transform(10, to_type='bool')

        with self.assertRaises(TypeError, msg='float should raise an error'):
            LookupHandler.transform(10.0, to_type='bool')

        with self.assertRaises(TypeError, msg='NoneType should raise an error'):
            LookupHandler.transform(None, to_type='bool')

    def test_transform_str_direct(self):
        """Test types that are directly transformed to strings."""
        self.assertEqual(LookupHandler.transform('test', 'str'), 'test')
        self.assertEqual(LookupHandler.transform({'key1': 'val1'}, 'str'),
                         json.dumps(json.dumps({'key1': 'val1'}, indent=0)))
        self.assertEqual(LookupHandler.transform(True, 'str'), '"True"')

    def test_transform_str_list(self):
        """Test list type joined to create string."""
        self.assertEqual(
            LookupHandler.transform(['val1', 'val2'], to_type='str'),
            'val1,val2'
        )
        self.assertEqual(
            LookupHandler.transform(set(['val', 'val']), to_type='str'),
            'val'
        )
        self.assertEqual(
            LookupHandler.transform(('val1', 'val2'), to_type='str'),
            'val1,val2'
        )

    def test_transform_str_list_delimiter(self):
        """Test list to string with a specified delimiter."""
        self.assertEqual(
            LookupHandler.transform(['val1', 'val2'], to_type='str',
                                    delimiter='|'),
            'val1|val2'
        )



"""Tests for runway.cfngin.lookups.handlers.envvar."""
import os
import unittest

from runway.cfngin.lookups.handlers.envvar import EnvvarLookup


class TestEnvVarHandler(unittest.TestCase):
    """Tests for runway.cfngin.lookups.handlers.envvar.EnvvarLookup."""

    def setUp(self):
        """Run before tests."""
        self.testkey = 'STACKER_ENVVAR_TESTCASE'
        self.invalidtestkey = 'STACKER_INVALID_ENVVAR_TESTCASE'
        self.testval = 'TestVal'
        os.environ[self.testkey] = self.testval

    def test_valid_envvar(self):
        """Test valid envvar."""
        value = EnvvarLookup.handle(self.testkey)
        self.assertEqual(value, self.testval)

    def test_invalid_envvar(self):
        """Test invalid envvar."""
        with self.assertRaises(ValueError):
            EnvvarLookup.handle(self.invalidtestkey)



"""Lookup to provide a default value."""
# pylint: disable=arguments-differ,unused-argument
from runway.lookups.handlers.base import LookupHandler


TYPE_NAME = "default"


class DefaultLookup(LookupHandler):
    """Lookup to provide a default value."""

    @classmethod
    def handle(cls, value, context=None, provider=None, **kwargs):
        """Use a value from the environment or fall back to a default value.

        Allows defaults to be set at the config file level.

        Args:
            value (str): Parameter(s) given to this lookup.
                ``<env_var>::<default value>``
            context (:class:`runway.cfngin.context.Context`): Context instance.
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance.

        Returns:
            str: Looked up value

        Example:
            ::

                Groups: ${default app_security_groups::sg-12345,sg-67890}

            If ``app_security_groups`` is defined in the environment, its
            defined value will be returned. Otherwise, ``sg-12345,sg-67890``
            will be thereturned value.

        """
        try:
            env_var_name, default_val = value.split("::", 1)
        except ValueError:
            raise ValueError("Invalid value for default: %s. Must be in "
                             "<env_var>::<default value> format." % value)

        if env_var_name in context.environment:
            return context.environment[env_var_name]
        return default_val



"""CFNgin test."""
# flake8: noqa
# pylint: disable=invalid-name
from os.path import basename

from integration_tests.test_cfngin.test_cfngin import Cfngin

FILE_BASENAME = '.'.join(basename(__file__).split('.')[:-1])


class TestSimpleDiff(Cfngin):
    """Test CFNgin diff.

    Requires valid AWS credentials.

    """

    REQUIRED_FIXTURE_FILES = [
        FILE_BASENAME + '.yaml'
    ]
    TEST_NAME = __name__

    def _build(self):
        """Execute and assert initial build."""
        self.set_environment('dev')
        code, _stdout, _stderr = self.runway_cmd('deploy')
        assert code == 0, 'exit code should be zero'

    def _diff(self):
        """Execute and assert second build with no changes."""
        self.set_environment('dev2')
        code, _stdout, stderr = self.runway_cmd('plan')
        assert code == 0, 'exit code should be zero'
        expected_lines = [
            '-InstanceType = m5.large',
            '+InstanceType = m5.xlarge',
            '- ResourceChange:',
            '    Action: Add',
            '    LogicalResourceId: VPC1',
            '    ResourceType: AWS::CloudFormation::WaitConditionHandle'
        ]
        for line in expected_lines:
            assert line in stderr, f'"{line}" missing from output'

    def run(self):
        """Run the test."""
        self.copy_fixtures()
        self._build()
        self._diff()

    def teardown(self):
        """Teardown any created resources and delete files."""
        self.runway_cmd('destroy')
        self.cleanup_fixtures()



"""Set package version."""
import sys

from . import cfngin, variables

sys.modules['stacker'] = cfngin  # shim to remove stacker dependency
sys.modules['stacker.variables'] = variables  # shim to support standard variables

__version__ = '1.5.0'



"""Static website module."""

import logging
import os
import sys
import tempfile
import warnings

import yaml

from . import RunwayModule
from .cloudformation import CloudFormation

LOGGER = logging.getLogger('runway')


def add_url_scheme(url):
    """Add the scheme to an existing url.

    Args:
        url (str): The current url
    """
    if url.startswith('https://') or url.startswith('http://'):
        return url
    newurl = 'https://%s' % url
    return newurl


class StaticSite(RunwayModule):
    """Static website Runway Module."""

    def __init__(self, context, path, options=None):
        """Initialize."""
        super(StaticSite, self).__init__(context, path, options)
        self.name = self.options.get('name', self.options.get('path'))
        self.user_options = self.options.get('options', {})
        self.parameters = self.options.get('parameters')
        # Memoize
        self.user_pool_id = ''
        self.region = self.context.env_region
        self._ensure_valid_environment_config()
        self._ensure_cloudfront_with_auth_at_edge()
        self._ensure_correct_region_with_auth_at_edge()

    def plan(self):
        """Create website CFN module and run stacker diff."""
        if self.parameters:
            self._setup_website_module(command='plan')
        else:
            LOGGER.info("Skipping staticsite plan of %s; no environment "
                        "config found for this environment/region",
                        self.options['path'])

    def deploy(self):
        """Create website CFN module and run stacker build."""
        if self.parameters:
            if self.parameters.get('staticsite_cf_disable', False) is False:
                msg = ("Please Note: Initial creation or updates to distribution settings "
                       "(e.g. url rewrites) will take quite a while (up to an hour). "
                       "Unless you receive an error your deployment is still running.")
                LOGGER.info(msg.upper())

                # Auth@Edge warning about subsequent deploys
                if self.parameters.get('staticsite_auth_at_edge', False) is True:
                    msg = ("PLEASE NOTE: A hook that is part of the dependencies stack of "
                           "the Auth@Edge static site deployment is designed to verify that "
                           "the correct Callback URLs are being used when a User Pool Client "
                           "already exists for the application. This ensures that there is no "
                           "interruption of service while the deployment reaches the stage "
                           "where the Callback URLs are updated to that of the Distribution. "
                           "Because of this you will receive a change set request on your first "
                           "subsequent deploy from the first, or any subsequent ones where the "
                           "alias domains have changed.")
                    LOGGER.info(msg)

            self._setup_website_module(command='deploy')
        else:
            LOGGER.info("Skipping staticsite deploy of %s; no environment "
                        "config found for this environment/region",
                        self.options['path'])

    def destroy(self):
        """Create website CFN module and run stacker destroy."""
        if self.parameters:
            self._setup_website_module(command='destroy')
        else:
            LOGGER.info("Skipping staticsite destroy of %s; no environment "
                        "config found for this environment/region",
                        self.options['path'])

    def _setup_website_module(self, command):
        """Create stacker configuration for website module."""
        module_dir = self._create_module_directory()
        self._create_dependencies_yaml(module_dir)
        self._create_staticsite_yaml(module_dir)

        cfn = CloudFormation(
            self.context,
            module_dir,
            {i: self.options[i] for i in self.options if i != 'class_path'}
        )
        getattr(cfn, command)()

    def _create_module_directory(self):
        module_dir = tempfile.mkdtemp()
        LOGGER.info("staticsite: Generating CloudFormation configuration for "
                    "module %s in %s",
                    self.name,
                    module_dir)
        return module_dir

    def _create_dependencies_yaml(self, module_dir):
        pre_build = []

        pre_destroy = [
            {'path': 'runway.hooks.cleanup_s3.purge_bucket',
             'required': True,
             'args': {
                 'bucket_rxref_lookup': "%s-dependencies::%s" % (self.name, i)  # noqa
             }} for i in ['AWSLogBucketName', 'ArtifactsBucketName']
        ]

        if self.parameters.get('staticsite_auth_at_edge', False):
            pre_build = [{
                'path': 'runway.hooks.staticsite.auth_at_edge.callback_url_retriever.get',
                'required': True,
                'data_key': 'aae_callback_url_retriever',
                'args': {
                    'user_pool_arn': self.parameters.get('staticsite_user_pool_arn', ''),
                    'stack_name': "${namespace}-%s-dependencies" % self.name
                }
            }]

            if self.parameters.get('staticsite_create_user_pool', False):
                pre_destroy.append({
                    'path': 'runway.hooks.staticsite.auth_at_edge.user_pool_id_retriever.get',
                    'required': True,
                    'data_key': 'aae_user_pool_id_retriever',
                    'args': self._get_user_pool_id_retriever_variables(),
                })

                pre_destroy.append({
                    'path': 'runway.hooks.staticsite.auth_at_edge.domain_updater.delete',
                    'required': True,
                    'data_key': 'aae_domain_updater',
                    'args': self._get_domain_updater_variables(),
                })

        with open(os.path.join(module_dir, '01-dependencies.yaml'), 'w') as output_stream:  # noqa
            yaml.dump(
                {'namespace': '${namespace}',
                 'cfngin_bucket': '',
                 'stacks': {
                     "%s-dependencies" % self.name: {
                         'class_path': 'runway.blueprints.staticsite.dependencies.Dependencies',
                         'variables': self._get_dependencies_variables()}},
                 'pre_build': pre_build,
                 'pre_destroy': pre_destroy},
                output_stream,
                default_flow_style=False
            )

    def _create_staticsite_yaml(self, module_dir):
        # Default parameter name matches build_staticsite hook
        hash_param = self.user_options.get('source_hashing', {}).get(
            'parameter',
            "${namespace}-%s-hash" % self.name
        )

        build_staticsite_args = self.options.copy() or {}
        build_staticsite_args['artifact_bucket_rxref_lookup'] = "%s-dependencies::ArtifactsBucketName" % self.name  # noqa pylint: disable=line-too-long
        build_staticsite_args['options']['namespace'] = '${namespace}'
        build_staticsite_args['options']['name'] = self.name
        build_staticsite_args['options']['path'] = os.path.join(
            os.path.realpath(self.context.env_root),
            self.path
        )

        site_stack_variables = self._get_site_stack_variables()

        class_path = 'staticsite.StaticSite'

        pre_build = [{'path': 'runway.hooks.staticsite.build_staticsite.build',
                      'required': True,
                      'data_key': 'staticsite',
                      'args': build_staticsite_args}]

        post_build = [{'path': 'runway.hooks.staticsite.upload_staticsite.sync',
                       'required': True,
                       'args': {
                           'bucket_output_lookup': '%s::BucketName' % self.name,
                           'website_url': '%s::BucketWebsiteURL' % self.name,
                           'cf_disabled': site_stack_variables['DisableCloudFront'],
                           'distributionid_output_lookup': '%s::CFDistributionId' % (self.name),
                           'distributiondomain_output_lookup': '%s::CFDistributionDomainName' % self.name}}]  # noqa pylint: disable=line-too-long

        pre_destroy = [{'path': 'runway.hooks.cleanup_s3.purge_bucket',
                        'required': True,
                        'args': {'bucket_rxref_lookup': "%s::BucketName" % self.name}}]

        post_destroy = [{'path': 'runway.hooks.cleanup_ssm.delete_param',
                         'args': {'parameter_name': hash_param}}]

        if self.parameters.get('staticsite_auth_at_edge', False):
            class_path = 'auth_at_edge.AuthAtEdge'

            pre_build.append({
                'path': 'runway.hooks.staticsite.auth_at_edge.user_pool_id_retriever.get',
                'required': True,
                'data_key': 'aae_user_pool_id_retriever',
                'args': self._get_user_pool_id_retriever_variables()
            })
            pre_build.append({
                'path': 'runway.hooks.staticsite.auth_at_edge.domain_updater.update',
                'required': True,
                'data_key': 'aae_domain_updater',
                'args': self._get_domain_updater_variables()
            })
            pre_build.append({
                'path': 'runway.hooks.staticsite.auth_at_edge.lambda_config.write',
                'required': True,
                'data_key': 'aae_lambda_config',
                'args': self._get_lambda_config_variables(
                    site_stack_variables
                )
            })
            post_build.insert(0, {
                'path': 'runway.hooks.staticsite.auth_at_edge.client_updater.update',
                'required': True,
                'data_key': 'client_updater',
                'args': self._get_client_updater_variables(
                    self.name,
                    site_stack_variables
                )
            })

        if self.parameters.get('staticsite_role_boundary_arn', False):
            site_stack_variables['RoleBoundaryArn'] = \
                self.parameters.pop('staticsite_role_boundary_arn')

        # If lambda_function_associations or custom_error_responses defined,
        # add to stack config
        for i in ['custom_error_responses', 'lambda_function_associations']:
            if self.parameters.get("staticsite_%s" % i):
                site_stack_variables[i] = self.parameters.pop("staticsite_%s" % i)

        with open(os.path.join(module_dir, '02-staticsite.yaml'), 'w') as output_stream:  # noqa
            yaml.dump(
                {'namespace': '${namespace}',
                 'cfngin_bucket': '',
                 'pre_build': pre_build,
                 'stacks': {
                     self.name: {
                         'class_path': 'runway.blueprints.staticsite.%s' % class_path,  # noqa
                         'variables': site_stack_variables}},
                 'post_build': post_build,
                 'pre_destroy': pre_destroy,
                 'post_destroy': post_destroy},
                output_stream,
                default_flow_style=False
            )

    def _get_site_stack_variables(self):
        site_stack_variables = {
            'Aliases': [],
            'DisableCloudFront': self.parameters.get('staticsite_cf_disable', False),
            'RewriteDirectoryIndex': self.parameters.get(
                'staticsite_rewrite_directory_index',
                ''
            ),
            'RedirectPathSignIn': '${default staticsite_redirect_path_sign_in::/parseauth}',
            'RedirectPathSignOut': '${default staticsite_redirect_path_sign_out::/}',
            'RedirectPathAuthRefresh':
                '${default staticsite_redirect_path_auth_refresh::/refreshauth}',
            'SignOutUrl': '${default staticsite_sign_out_url::/signout}',
            'WAFWebACL': self.parameters.get('staticsite_web_acl', '')
        }

        if self.parameters.get('staticsite_aliases'):
            site_stack_variables['Aliases'] = self.parameters.get('staticsite_aliases').split(',')

        if self.parameters.get('staticsite_acmcert_arn'):
            site_stack_variables['AcmCertificateArn'] = \
                self.parameters['staticsite_acmcert_arn']

        if self.parameters.get('staticsite_acmcert_ssm_param'):
            dep_msg = ('Use of the "staticsite_acmcert_ssm_param" option has '
                       'been deprecated. The "staticsite_acmcert_arn" option '
                       'with an "ssm" lookup should be used instead.')
            warnings.warn(dep_msg, DeprecationWarning)
            LOGGER.warning(dep_msg)
            site_stack_variables['AcmCertificateArn'] = '${ssmstore ${staticsite_acmcert_ssm_param}}'  # noqa pylint: disable=line-too-long

        if self.parameters.get('staticsite_enable_cf_logging', True):
            site_stack_variables['LogBucketName'] = "${rxref %s-dependencies::AWSLogBucketName}" % self.name  # noqa pylint: disable=line-too-long

        if self.parameters.get('staticsite_auth_at_edge', False):
            self._ensure_auth_at_edge_requirements()
            site_stack_variables['UserPoolArn'] = self.parameters.get(
                'staticsite_user_pool_arn'
            )
            site_stack_variables['NonSPAMode'] = self.parameters.get(
                'staticsite_non_spa',
                False
            )
            site_stack_variables['HttpHeaders'] = self._get_http_headers()
            site_stack_variables['CookieSettings'] = self._get_cookie_settings()
            site_stack_variables['OAuthScopes'] = self._get_oauth_scopes()
            # pylint: disable=line-too-long
            site_stack_variables['SupportedIdentityProviders'] = self._get_supported_identity_providers()  # noqa
        else:
            # If lambda_function_associations or custom_error_responses defined,
            # add to stack config. Only if not using Auth@Edge
            for i in ['custom_error_responses', 'lambda_function_associations']:
                if self.parameters.get("staticsite_%s" % i):
                    site_stack_variables[i] = self.parameters.get("staticsite_%s" % i)
                    self.parameters.pop("staticsite_%s" % i)

        return site_stack_variables

    def _get_cookie_settings(self):
        """Retrieve the cookie settings from the variables or return the default."""
        if self.parameters.get('staticsite_cookie_settings'):
            return self.parameters.get('staticsite_cookie_settings')
        return {
            "idToken": "Path=/; Secure; SameSite=Lax",
            "accessToken": "Path=/; Secure; SameSite=Lax",
            "refreshToken": "Path=/; Secure; SameSite=Lax",
            "nonce": "Path=/; Secure; HttpOnly; Max-Age=1800; SameSite=Lax",
        }

    def _get_http_headers(self):
        """Retrieve the http headers from the variables or return the default."""
        if self.parameters.get('staticsite_http_headers'):
            return self.parameters.get('staticsite_http_headers')
        return {
            "Content-Security-Policy": "default-src https: 'unsafe-eval' 'unsafe-inline'; "
                                       # pylint: disable=line-too-long
                                       "font-src 'self' 'unsafe-inline' 'unsafe-eval' data: https:; "  # noqa
                                       "object-src 'none'; "
                                       # pylint: disable=line-too-long
                                       "connect-src 'self' https://*.amazonaws.com https://*.amazoncognito.com",  # noqa
            "Strict-Transport-Security": "max-age=31536000; "
                                         "includeSubdomains; "
                                         "preload",
            "Referrer-Policy": "same-origin",
            "X-XSS-Protection": "1; mode=block",
            "X-Frame-Options": "DENY",
            "X-Content-Type-Options": "nosniff",
        }

    def _get_oauth_scopes(self):
        """Retrieve the oauth scopes from the variables or return the default."""
        if self.parameters.get('staticsite_oauth_scopes'):
            return self.parameters.get('staticsite_oauth_scopes')
        return [
            'phone',
            'email',
            'profile',
            'openid',
            'aws.cognito.signin.user.admin'
        ]

    def _get_supported_identity_providers(self):
        providers = self.parameters.get('staticsite_supported_identity_providers')
        if providers:
            return [provider.strip() for provider in providers.split(',')]
        return ['COGNITO']

    def _get_dependencies_variables(self):
        variables = {'OAuthScopes': self._get_oauth_scopes()}
        if self.parameters.get('staticsite_auth_at_edge', False):
            self._ensure_auth_at_edge_requirements()

            variables.update({
                'AuthAtEdge': self.parameters.get(
                    'staticsite_auth_at_edge',
                    False
                )
            })

            if self.parameters.get('staticsite_create_user_pool', False):
                variables.update({
                    'CreateUserPool': self.parameters.get('staticsite_create_user_pool', False)
                })

        return variables

    def _get_user_pool_id_retriever_variables(self):
        args = {
            'user_pool_arn': self.parameters.get('staticsite_user_pool_arn', ''),
        }

        if self.parameters.get('staticsite_create_user_pool', False):
            args['created_user_pool_id'] = \
                '${rxref %s-dependencies::AuthAtEdgeUserPoolId}' % (self.name)

        return args

    def _get_domain_updater_variables(self):
        return {
            'client_id_output_lookup': "%s-dependencies::AuthAtEdgeClient" % self.name,  # noqa pylint: disable=line-too-long
            'client_id': "${rxref %s-dependencies::AuthAtEdgeClient}" % self.name,
        }

    def _get_lambda_config_variables(self, site_stack_variables):
        return {
            'client_id': "${rxref %s-dependencies::AuthAtEdgeClient}" % self.name,  # noqa pylint: disable=line-too-long
            'bucket': "${rxref %s-dependencies::ArtifactsBucketName}" % self.name,
            'cookie_settings': site_stack_variables['CookieSettings'],
            'http_headers': site_stack_variables['HttpHeaders'],
            'oauth_scopes': site_stack_variables['OAuthScopes'],
            'redirect_path_refresh': site_stack_variables['RedirectPathAuthRefresh'],
            'redirect_path_sign_in': site_stack_variables['RedirectPathSignIn'],
            'redirect_path_sign_out': site_stack_variables['RedirectPathSignOut'],
        }

    def _get_client_updater_variables(self, name, site_stack_variables):
        aliases = [add_url_scheme(x) for x in site_stack_variables['Aliases']]
        return {
            'alternate_domains': aliases,
            'client_id': "${rxref %s-dependencies::AuthAtEdgeClient}" % self.name,
            'distribution_domain': '${rxref %s::CFDistributionDomainName}' % name,
            'oauth_scopes': site_stack_variables['OAuthScopes'],
            'redirect_path_sign_in': site_stack_variables['RedirectPathSignIn'],
            'redirect_path_sign_out': site_stack_variables['RedirectPathSignOut'],
            'supported_identity_providers': site_stack_variables['SupportedIdentityProviders'],
        }

    def _ensure_auth_at_edge_requirements(self):
        if not self.parameters.get('staticsite_user_pool_arn') and \
           not self.parameters.get('staticsite_create_user_pool'):
            LOGGER.fatal("A Cognito UserPool ARN is required with Auth@Edge. "
                         "You can supply your own or have one created.")
            sys.exit(1)

    def _ensure_correct_region_with_auth_at_edge(self):
        """Exit if not in the us-east-1 region and deploying to Auth@Edge.

        Lambda@Edge is only available within the us-east-1 region.
        """
        if self.parameters.get('staticsite_auth_at_edge', False) and self.region != 'us-east-1':
            LOGGER.fatal("Auth@Edge must be deployed in us-east-1.")
            sys.exit(1)

    def _ensure_cloudfront_with_auth_at_edge(self):
        """Exit if both the Auth@Edge and CloudFront disablement are true."""
        if self.parameters.get('staticsite_cf_disable', False) and \
           self.parameters.get('staticsite_auth_at_edge', False):
            LOGGER.fatal("CloudFront cannot be disabled when using Auth@Edge")
            sys.exit(1)

    def _ensure_valid_environment_config(self):
        """Exit if config is invalid."""
        if not self.parameters.get('namespace'):
            LOGGER.fatal("staticsite: module %s's environment configuration is "
                         "missing a namespace definition!",
                         self.name)
            sys.exit(1)



"""Sign the user out of Cognito and remove all Cookie Headers."""

import logging
from urllib.parse import urlencode  # pylint: disable=no-name-in-module,import-error

from shared import extract_and_parse_cookies, get_config, get_cookie_headers  # noqa pylint: disable=import-error

LOGGER = logging.getLogger(__name__)
CONFIG = get_config()


def handler(event, _context):
    """Handle the signout event."""
    request = event['Records'][0]['cf']['request']
    domain_name = request['headers']['host'][0]['value']
    extracted = extract_and_parse_cookies(
        request['headers'],
        CONFIG.get('client_id')
    )

    if not extracted.get('idToken'):
        return {
            'body': "Bad Request",
            'status': "400",
            'statusDescription': "Bad Request",
            'headers': CONFIG.get('cloud_front_headers')
        }

    tokens = {
        'id_token': extracted.get('idToken'),
        'access_token': extracted.get('accessToken'),
        'refresh_token': extracted.get('refreshToken'),
    }
    query_string = {
        'logout_uri': 'https://%s%s' % (domain_name, CONFIG.get('redirect_path_sign_out')),
        'client_id': CONFIG.get('client_id')
    }

    headers = {
        # Redirect the user to logout
        'location': [
            {
                'key': 'location',
                'value': 'https://%s/logout?%s' % (
                    CONFIG.get('cognito_auth_domain'),
                    urlencode(query_string)
                )
            }
        ],
        'set-cookie': get_cookie_headers(
            CONFIG.get("client_id"),
            CONFIG.get("oauth_scopes"),
            tokens,
            domain_name,
            CONFIG.get('cookie_settings'),
            # Make sure we expire all the tokens during retrieval
            expire_all_tokens=True
        )
    }
    headers.update(CONFIG.get('cloud_front_headers'))

    return {
        'status': '307',
        'statusDescription': 'Temporary Redirect',
        'headers': headers
    }



"""AWS Lambda hook."""
import hashlib
import logging
import os
import stat
import subprocess
import sys
# https://github.com/PyCQA/pylint/issues/2955
from distutils.util import strtobool  # pylint: disable=E
from io import BytesIO as StringIO
from shutil import copyfile
from types import GeneratorType
from zipfile import ZIP_DEFLATED, ZipFile

import botocore
import docker
import formic
from six import string_types
from troposphere.awslambda import Code

from ..exceptions import (InvalidDockerizePipConfiguration, PipenvError,
                          PipError)
from ..session_cache import get_session
from ..util import ensure_s3_bucket

if sys.version_info[0] < 3:
    from backports import tempfile  # pylint: disable=E
else:
    import tempfile

# mask to retrieve only UNIX file permissions from the external attributes
# field of a ZIP entry.
ZIP_PERMS_MASK = (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO) << 16

LOGGER = logging.getLogger(__name__)

# list from python tags of https://hub.docker.com/r/lambci/lambda/tags
SUPPORTED_RUNTIMES = [
    # Python 2.7 reached end-of-life on January 1st, 2020.
    # However, the Python 2.7 runtime is still supported and is not scheduled
    # to be deprecated at this time.
    # https://docs.aws.amazon.com/lambda/latest/dg/runtime-support-policy.html
    'python2.7',
    'python3.6',
    'python3.7',
    'python3.8'
]


def copydir(source, destination, includes, excludes=None,
            follow_symlinks=False):
    """Extend the functionality of shutil.

    Correctly copies files and directories in a source directory.

    Args:
        source (str): Source directory.
        destination (str): Destination directory.
        includes (List[str]): Glob patterns for files to include.
        includes (List[str]): Glob patterns for files to exclude.
        follow_symlinks (bool): If true, symlinks will be included in the
            resulting zip file.

    """
    files = _find_files(source, includes, excludes, follow_symlinks)

    def _mkdir(dir_name):
        """Recursively create directories."""
        parent = os.path.dirname(dir_name)
        if not os.path.isdir(parent):
            _mkdir(parent)
        LOGGER.debug('lambda.copydir: Creating directory: %s', dir_name)
        os.mkdir(dir_name)

    for file_name in files:
        src = os.path.join(source, file_name)
        dest = os.path.join(destination, file_name)
        try:
            LOGGER.debug('lambda.copydir: Copying file "%s" to "%s"',
                         src, dest)
            copyfile(src, dest)
        # python2 raises an IOError here
        except (IOError, OSError):
            _mkdir(os.path.dirname(dest))
            copyfile(src, dest)


def find_requirements(root):
    """Identify Python requirement files.

    Args:
        root (str): Path that should be searched for files.

    Returns:
        Optional[Dict[str, bool]]: Name of supported requirements file and
        wether it was found. If none are found, ``None`` is returned.

    """
    findings = {}

    for file_name in ['requirements.txt',
                      'Pipfile',
                      'Pipfile.lock']:
        findings[file_name] = os.path.isfile(os.path.join(root, file_name))

    if not sum(findings.values()):
        return None
    return findings


def should_use_docker(dockerize_pip=None):
    """Assess if Docker should be used based on the value of args.

    Args:
        dockerize_pip (Union[bool, None, str]): Value to assess if Docker
            should be used for pip.

    Returns:
        bool

    """
    if isinstance(dockerize_pip, bool):
        return dockerize_pip

    if isinstance(dockerize_pip, str):
        if dockerize_pip == 'non-linux' and \
                not sys.platform.startswith('linux'):
            return True
        try:
            return strtobool(dockerize_pip)
        except ValueError:
            pass
    return False


def _zip_files(files, root):
    """Generate a ZIP file in-memory from a list of files.

    Files will be stored in the archive with relative names, and have their
    UNIX permissions forced to 755 or 644 (depending on whether they are
    user-executable in the source filesystem).

    Args:
        files (List[str]): file names to add to the archive, relative to
            ``root``.
        root (str): base directory to retrieve files from.

    Returns:
        Tuple[str, str]: Content of the ZIP file as a byte string and
        calculated hash of all the files

    """
    zip_data = StringIO()
    if isinstance(files, GeneratorType):
        # if file list is a generator, save the contents so it can be reused
        # since generators are empty after the first iteration and cannot be
        # rewound.
        LOGGER.debug('lambda: Converting file generater to list for reuse...')
        files = list(files)
    with ZipFile(zip_data, 'w', ZIP_DEFLATED) as zip_file:
        for file_name in files:
            zip_file.write(os.path.join(root, file_name), file_name)

        # Fix file permissions to avoid any issues - only care whether a file
        # is executable or not, choosing between modes 755 and 644 accordingly.
        for zip_entry in zip_file.filelist:
            perms = (zip_entry.external_attr & ZIP_PERMS_MASK) >> 16
            if perms & stat.S_IXUSR != 0:
                new_perms = 0o755
            else:
                new_perms = 0o644

            if new_perms != perms:
                LOGGER.debug("lambda: fixing perms: %s: %o => %o",
                             zip_entry.filename, perms, new_perms)
                new_attr = (
                    (zip_entry.external_attr & ~ZIP_PERMS_MASK) | (new_perms << 16)
                )
                zip_entry.external_attr = new_attr

    contents = zip_data.getvalue()
    zip_data.close()
    content_hash = _calculate_hash(files, root)

    return contents, content_hash


def _calculate_hash(files, root):
    """Return a hash of all of the given files at the given root.

    Args:
        files (list[str]): file names to include in the hash calculation,
            relative to ``root``.
        root (str): base directory to analyze files in.

    Returns:
        str: A hash of the hashes of the given files.

    """
    file_hash = hashlib.md5()
    for file_name in sorted(files):
        file_path = os.path.join(root, file_name)
        file_hash.update((file_name + "\0").encode())
        with open(file_path, "rb") as file_:
            for chunk in iter(lambda: file_.read(4096), ""):  # pylint: disable=W
                if not chunk:
                    break
                file_hash.update(chunk)
            file_hash.update("\0".encode())

    return file_hash.hexdigest()


def _find_files(root, includes, excludes=None, follow_symlinks=False):
    """List files inside a directory based on include and exclude rules.

    This is a more advanced version of `glob.glob`, that accepts multiple
    complex patterns.

    Args:
        root (str): base directory to list files from.
        includes (List[str]): inclusion patterns. Only files matching those
            patterns will be included in the result.
        excludes (List[str]): exclusion patterns. Files matching those
            patterns will be excluded from the result. Exclusions take
            precedence over inclusions.
        follow_symlinks (bool): If true, symlinks will be included in the
            resulting zip file

    Yields:
        str: a file name relative to the root.

    Note:
        Documentation for the patterns can be found at
        http://www.aviser.asia/formic/doc/index.html

    """
    root = os.path.abspath(root)
    file_set = formic.FileSet(
        directory=root, include=includes,
        exclude=excludes, symlinks=follow_symlinks,
    )

    for filename in file_set.qualified_files(absolute=False):
        yield filename


def _zip_from_file_patterns(root, includes, excludes, follow_symlinks):
    """Generate a ZIP file in-memory from file search patterns.

    Args:
        root (str): Base directory to list files from.
        includes (List[str]): Inclusion patterns. Only files  matching those
            patterns will be included in the result.
        excludes (List[str]): Exclusion patterns. Files matching those
            patterns will be excluded from the result. Exclusions take
            precedence over inclusions.
        follow_symlinks (bool): If true, symlinks will be included in the
            resulting zip file

    See Also:
        :func:`_zip_files`, :func:`_find_files`.

    Raises:
        RuntimeError: when the generated archive would be empty.

    """
    LOGGER.info('lambda: base directory: %s', root)

    files = list(_find_files(root, includes, excludes, follow_symlinks))
    if not files:
        raise RuntimeError('Empty list of files for Lambda payload. Check '
                           'your include/exclude options for errors.')

    LOGGER.info('lambda: adding %d files:', len(files))

    for file_name in files:
        LOGGER.debug('lambda: + %s', file_name)

    return _zip_files(files, root)


def handle_requirements(package_root, dest_path, requirements,
                        pipenv_timeout=300, python_path=None,
                        use_pipenv=False):
    """Use the correct requirements file.

    Args:
        package_root (str): Base directory containing a requirements file.
        dest_path (str): Where to output the requirements file if one needs
            to be created.
        requirements (Dict[str, bool]): Map of requirement file names and
            wether they exist.
        pipenv_timeout (int): Seconds to wait for a subprocess to complete.
        python_path (Optional[str]): Explicit python interpreter to be used.
            Requirement file generators must be installed and executable using
            ``-m`` if provided.
        use_pipenv (bool): Explicitly use pipenv to export a Pipfile as
            requirements.txt.

    Returns:
        str: Path to the final requirements.txt

    Raises:
        NotImplementedError: When a requirements file is not found. This
            should never be encountered but is included just in case.

    """
    if use_pipenv:
        LOGGER.info('lambda: explicitly using pipenv')
        return _handle_use_pipenv(package_root=package_root,
                                  dest_path=dest_path, python_path=python_path,
                                  timeout=pipenv_timeout)
    if requirements['requirements.txt']:
        LOGGER.info('lambda: using requirements.txt for dependencies')
        return os.path.join(dest_path, 'requirements.txt')
    if requirements['Pipfile'] or requirements['Pipfile.lock']:
        LOGGER.info('lambda: using pipenv for dependencies')
        return _handle_use_pipenv(package_root=package_root,
                                  dest_path=dest_path, python_path=python_path,
                                  timeout=pipenv_timeout)
    # This point should never be reached under normal operation since a
    # requirements file of some sort must have been found in another step
    # of the process but just in case it does happen, raise an error.
    raise NotImplementedError('Unable to handle missing requirements file.')


def _handle_use_pipenv(package_root, dest_path, python_path=None, timeout=300):
    """Create requirements file from Pipfile.

    Args:
        package_root (str): Base directory to generate requirements from.
        dest_path (str): Where to output the requirements file.
        python_path (Optional[str]): Explicit python interpreter to be used.
            pipenv must be installed and executable using ``-m`` if provided.
        timeout (int): Seconds to wait for process to complete.

    Raises:
        PipenvError: Non-zero exit code returned by pipenv process.

    """
    LOGGER.info('lambda.pipenv: Creating requirements.txt from Pipfile...')
    req_path = os.path.join(dest_path, 'requirements.txt')
    cmd = ['pipenv', 'lock', '--requirements', '--keep-outdated']
    if python_path:
        cmd.insert(0, python_path)
        cmd.insert(1, '-m')
    with open(req_path, 'w') as requirements:
        pipenv_process = subprocess.Popen(cmd, cwd=package_root,
                                          stdout=requirements,
                                          stderr=subprocess.PIPE)
        if int(sys.version[0]) > 2:
            _stdout, stderr = pipenv_process.communicate(timeout=timeout)
        else:
            _stdout, stderr = pipenv_process.communicate()
        if pipenv_process.returncode == 0:
            return req_path
        if int(sys.version[0]) > 2:
            stderr = stderr.decode('UTF-8')
        LOGGER.error('lambda.pipenv: "%s" failed with the following '
                     'output:\n%s', ' '.join(cmd), stderr)
        raise PipenvError


def dockerized_pip(work_dir, client=None, runtime=None, docker_file=None,
                   docker_image=None, **_kwargs):
    """Run pip with docker.

    Args:
        work_dir (str): Work directory for docker.
        client (Optional[docker.DockerClient]): Custom docker client.
        runtime (Optional[str]): Lambda runtime. Must provide one of
            ``runtime``, ``docker_file``, or ``docker_image``.
        docker_file (Optional[str]): Path to a Dockerfile to build an image.
            Must provide one of ``runtime``, ``docker_file``, or
            ``docker_image``.
        docker_image (Optional[str]): Local or remote docker image to use.
            Must provide one of ``runtime``, ``docker_file``, or
            ``docker_image``.
        kwargs (Any): Advanced options for docker. See source code to
            determine what is supported.

    Returns:
        Tuple[str, str]: Content of the ZIP file as a byte string and
        calculated hash of all the files

    """
    # TODO use kwargs to pass args to docker for advanced config
    if bool(docker_file) + bool(docker_image) + bool(runtime) != 1:
        # exactly one of these is needed. converting to bool will give us a
        # 'False' (0) for 'None' and 'True' (1) for anything else.
        raise InvalidDockerizePipConfiguration(
            'exactly only one of [docker_file, docker_file, runtime] must be '
            'provided'
        )

    if not client:
        client = docker.from_env()

    if docker_file:
        if not os.path.isfile(docker_file):
            raise ValueError('could not find docker_file "%s"' % docker_file)
        LOGGER.info('lambda.docker: Building docker image from "%s".',
                    docker_file)
        response = client.images.build(
            path=os.path.dirname(docker_file),
            dockerfile=os.path.basename(docker_file),
            forcerm=True
        )
        # the response can be either a tuple of (Image, Generator[Dict[str, str]])
        # or just Image depending on API version.
        if isinstance(response, tuple):
            docker_image = response[0].id
            for log_msg in response[1]:
                if log_msg.get('stream'):
                    LOGGER.info(log_msg['stream'].strip('\n'))
        else:
            docker_image = response.id
        LOGGER.info('lambda.docker: Docker image "%s" created.', docker_image)
    if runtime:
        if runtime not in SUPPORTED_RUNTIMES:
            raise ValueError('invalid runtime "{}" must be one of {}'.format(
                runtime, str(SUPPORTED_RUNTIMES)
            ))
        docker_image = 'lambci/lambda:build-%s' % runtime
        LOGGER.debug('lambda.docker: Selected docker image "%s" based on '
                     'provided runtime', docker_image)

    if sys.platform.lower() == 'win32':
        LOGGER.debug('lambda.docker: Formatted docker mount path for Windows')
        work_dir = work_dir.replace('\\', '/')

    work_dir_mount = docker.types.Mount(target='/var/task',
                                        source=work_dir,
                                        type='bind')
    pip_cmd = (
        'python -m pip install -t /var/task -r /var/task/requirements.txt'
    )

    LOGGER.info('lambda.docker: Using docker image "%s" to build deployment '
                'package...', docker_image)

    container = client.containers.run(image=docker_image,
                                      command=['/bin/sh', '-c', pip_cmd],
                                      auto_remove=True,
                                      detach=True,
                                      mounts=[work_dir_mount])

    # 'stream' creates a blocking generator that allows for real-time logs.
    # this loop ends when the container 'auto_remove's itself.
    for log in container.logs(stdout=True,
                              stderr=True,
                              stream=True,
                              tail=0):
        # without strip there are a bunch blank lines in the output
        LOGGER.info('lambda.docker: %s', log.decode().strip())


def _zip_package(package_root, includes, excludes=None, dockerize_pip=False,
                 follow_symlinks=False, python_path=None,
                 requirements_files=None, use_pipenv=False, **kwargs):
    """Create zip file in memory with package dependencies.

    Args:
        package_root (str): Base directory to copy files from.
        includes (List[str]): Inclusion patterns. Only files  matching those
            patterns will be included in the result.
        excludes (List[str]): Exclusion patterns. Files matching those
            patterns will be excluded from the result. Exclusions take
            precedence over inclusions.
        dockerize_pip (Union[bool, str]): Whether to use docker or under what
            conditions docker will be used to run ``pip``.
        follow_symlinks (bool): If true, symlinks will be included in the
            resulting zip file.
        python_path (Optional[str]): Explicit python interpreter to be used.
            pipenv must be installed and executable using ``-m`` if provided.
        requirements_files (Dict[str, bool]): Map of requirement file names and
            wether they exist.
        use_pipenv (bool): Wether to use pipenv to export a Pipfile as
            requirements.txt.
        kwargs (Any): Advanced options for subprocess and docker. See source
            code to determine what is supported.

    Returns:
        Tuple[str, str]: Content of the ZIP file as a byte string and
        calculated hash of all the files

    """
    kwargs.setdefault('pipenv_timeout', 300)

    temp_root = os.path.join(os.path.expanduser('~'), '.runway_cache')
    if not os.path.isdir(temp_root):
        os.makedirs(temp_root)

    # exclude potential virtual environments in the package
    excludes.append('.venv/')

    with tempfile.TemporaryDirectory(
            prefix='cfngin', dir=temp_root
    ) as tmpdir:
        tmp_req = os.path.join(tmpdir, 'requirements.txt')
        copydir(package_root, tmpdir, includes, excludes, follow_symlinks)
        tmp_req = handle_requirements(package_root=package_root,
                                      dest_path=tmpdir,
                                      requirements=requirements_files,
                                      python_path=python_path,
                                      use_pipenv=use_pipenv,
                                      pipenv_timeout=kwargs['pipenv_timeout'])

        if should_use_docker(dockerize_pip):
            dockerized_pip(tmpdir, **kwargs)
        else:
            pip_cmd = [python_path or sys.executable, '-m',
                       'pip', 'install',
                       '-t', tmpdir,
                       '-r', tmp_req]
            pip_proc = subprocess.Popen(pip_cmd,
                                        cwd=tmpdir, stdout=subprocess.PIPE,
                                        stderr=subprocess.PIPE)
            if int(sys.version[0]) > 2:
                _stdout, stderr = pip_proc.communicate(timeout=kwargs.get(
                    'pipenv_timeout', 900
                ))
            else:
                _stdout, stderr = pip_proc.communicate()
            if pip_proc.returncode != 0:
                if int(sys.version[0]) > 2:
                    stderr = stderr.decode('UTF-8')
                LOGGER.error('"%s" failed with the following output:\n%s',
                             ' '.join(pip_cmd), stderr)
                raise PipError

        req_files = _find_files(tmpdir, includes='**', follow_symlinks=False)
        return _zip_files(req_files, tmpdir)


def _head_object(s3_conn, bucket, key):
    """Retrieve information about an object in S3 if it exists.

    Args:
        s3_conn (botocore.client.S3): S3 connection to use for operations.
        bucket (str): name of the bucket containing the key.
        key (str): name of the key to lookup.

    Returns:
        Dict[str, Any]: S3 object information, or None if the object does not
        exist. See the AWS documentation for explanation of the contents.

    Raises:
        botocore.exceptions.ClientError: any error from boto3 other than key
            not found is passed through.

    """
    try:
        return s3_conn.head_object(Bucket=bucket, Key=key)
    except botocore.exceptions.ClientError as err:
        if err.response['Error']['Code'] == '404':
            return None
        raise


def _upload_code(s3_conn, bucket, prefix, name, contents, content_hash,
                 payload_acl):
    """Upload a ZIP file to S3 for use by Lambda.

    The key used for the upload will be unique based on the checksum of the
    contents. No changes will be made if the contents in S3 already match the
    expected contents.

    Args:
        s3_conn (botocore.client.S3): S3 connection to use for operations.
        bucket (str): name of the bucket to create.
        prefix (str): S3 prefix to prepend to the constructed key name for
            the uploaded file
        name (str): desired name of the Lambda function. Will be used to
            construct a key name for the uploaded file.
        contents (str): byte string with the content of the file upload.
        content_hash (str): md5 hash of the contents to be uploaded.
        payload_acl (str): The canned S3 object ACL to be applied to the
            uploaded payload

    Returns:
        troposphere.awslambda.Code: CloudFormation Lambda Code object,
        pointing to the uploaded payload in S3.

    Raises:
        botocore.exceptions.ClientError: any error from boto3 is passed
            through.

    """
    LOGGER.debug('lambda: ZIP hash: %s', content_hash)
    key = '{}lambda-{}-{}.zip'.format(prefix, name, content_hash)

    if _head_object(s3_conn, bucket, key):
        LOGGER.info('lambda: object %s already exists, not uploading', key)
    else:
        LOGGER.info('lambda: uploading object %s', key)
        s3_conn.put_object(Bucket=bucket, Key=key, Body=contents,
                           ContentType='application/zip',
                           ACL=payload_acl)

    return Code(S3Bucket=bucket, S3Key=key)


def _check_pattern_list(patterns, key, default=None):
    """Validate file search patterns from user configuration.

    Acceptable input is a string (which will be converted to a singleton list),
    a list of strings, or anything falsy (such as None or an empty dictionary).
    Empty or unset input will be converted to a default.

    Args:
        patterns: Input from user configuration (YAML).
        key (str): Name of the configuration key the input came from,
            used for error display purposes.

    Keyword Args:
        default: Value to return in case the input is empty or unset.

    Returns:
        List[str]: Validated list of patterns.

    Raises:
        ValueError: If the input is unacceptable.

    """
    if not patterns:
        return default

    if isinstance(patterns, string_types):
        return [patterns]

    if isinstance(patterns, list):
        if all(isinstance(p, string_types) for p in patterns):
            return patterns

    raise ValueError("Invalid file patterns in key '{}': must be a string or "
                     'list of strings'.format(key))


def _upload_function(s3_conn, bucket, prefix, name, options, follow_symlinks,
                     payload_acl, sys_path):
    """Build a Lambda payload from user configuration and uploads it to S3.

    Args:
        s3_conn (botocore.client.S3): S3 connection to use for operations.
        bucket (str): name of the bucket to upload to.
        prefix (str): S3 prefix to prepend to the constructed key name for
            the uploaded file
        name (str): Desired name of the Lambda function. Will be used to
            construct a key name for the uploaded file.
        options (Dict[str, Any]): Configuration for how to build the payload.
            Consists of the following keys:
                **path**:
                    Base path to retrieve files from (mandatory). If not
                    absolute, it will be interpreted as relative to the CFNgin
                    configuration file directory, then converted to an absolute
                    path. See :func:`runway.cfngin.util.get_config_directory`.
                **include**:
                    File patterns to include in the payload (optional).
                **exclude**:
                    File patterns to exclude from the payload (optional).
        follow_symlinks (bool): If true, symlinks will be included in the
            resulting zip file
        payload_acl (str): The canned S3 object ACL to be applied to the
            uploaded payload
        sys_path (str): Path that all actions are relative to.

    Returns:
        troposphere.awslambda.Code: CloudFormation AWS Lambda Code object,
        pointing to the uploaded object in S3.

    Raises:
        ValueError: If any configuration is invalid.
        botocore.exceptions.ClientError: Any error from boto3 is passed
            through.

    """
    try:
        root = os.path.expanduser(options['path'])
    except KeyError as err:
        raise ValueError(
            "missing required property '{}' in function '{}'".format(
                err.args[0], name))

    includes = _check_pattern_list(options.get('include'), 'include',
                                   default=['**'])
    excludes = _check_pattern_list(options.get('exclude'), 'exclude',
                                   default=[])

    LOGGER.debug('lambda: processing function %s', name)

    # os.path.join will ignore other parameters if the right-most one is an
    # absolute path, which is exactly what we want.
    if not os.path.isabs(root):
        root = os.path.abspath(os.path.join(sys_path, root))
    requirements_files = find_requirements(root)
    if requirements_files:
        zip_contents, content_hash = _zip_package(
            root,
            includes=includes,
            excludes=excludes,
            follow_symlinks=follow_symlinks,
            requirements_files=requirements_files,
            **options
        )
    else:
        zip_contents, content_hash = _zip_from_file_patterns(root,
                                                             includes,
                                                             excludes,
                                                             follow_symlinks)

    return _upload_code(s3_conn, bucket, prefix, name, zip_contents,
                        content_hash, payload_acl)


def select_bucket_region(custom_bucket, hook_region, cfngin_bucket_region,
                         provider_region):
    """Return the appropriate region to use when uploading functions.

    Select the appropriate region for the bucket where lambdas are uploaded in.

    Args:
        custom_bucket (Optional[str]): The custom bucket name provided by the
            `bucket` kwarg of the aws_lambda hook, if provided.
        hook_region (str): The contents of the `bucket_region` argument to
            the hook.
        cfngin_bucket_region (str): The contents of the
            ``cfngin_bucket_region`` global setting.
        provider_region (str): The region being used by the provider.

    Returns:
        str: The appropriate region string.

    """
    region = None
    if custom_bucket:
        region = hook_region
    else:
        region = cfngin_bucket_region
    return region or provider_region


def upload_lambda_functions(context, provider, **kwargs):
    """Build Lambda payloads from user configuration and upload them to S3.

    Constructs ZIP archives containing files matching specified patterns for
    each function, uploads the result to Amazon S3, then stores objects (of
    type :class:`troposphere.awslambda.Code`) in the context's hook data,
    ready to be referenced in blueprints.

    Configuration consists of some global options, and a dictionary of function
    specifications. In the specifications, each key indicating the name of the
    function (used for generating names for artifacts), and the value
    determines what files to include in the ZIP (see more details below).

    Payloads are uploaded to either a custom bucket or the CFNgin default
    bucket, with the key containing it's checksum, to allow repeated uploads
    to be skipped in subsequent runs.

    The configuration settings are documented as keyword arguments below.

    Args:
        provider (:class:`runway.cfngin.providers.base.BaseProvider`): Provider
            instance. (passed in by CFNgin)
        context (:class:`runway.cfngin.context.Context`): Context instance.
            (passed in by CFNgin)

    Keyword Args:
        bucket (Optional[str]): Custom bucket to upload functions to.
            Omitting it will cause the default CFNgin bucket to be used.
        bucket_region (Optional[str]): The region in which the bucket should
            exist. If not given, the region will be either be that of the
            global ``cfngin_bucket_region`` setting, or else the region in
            use by the provider.
        prefix (Optional[str]): S3 key prefix to prepend to the uploaded
            zip name.
        follow_symlinks (Optional[bool]): Will determine if symlinks should
            be followed and included with the zip artifact. (*default:*
            ``False``)
        payload_acl (Optional[str]): The canned S3 object ACL to be applied
            to the uploaded payload. (*default: private*)
        functions (Dict[str, Any]): Configurations of desired payloads to
            build. Keys correspond to function names, used to derive key
            names for the payload. Each value should itself be a dictionary,
            with the following data:

            **docker_file (Optional[str])**
                Path to a local DockerFile that will be built and used for
                ``dockerize_pip``. Must provide exactly one of ``docker_file``,
                ``docker_image``, or ``runtime``.

            **docker_image (Optional[str])**
                Custom Docker image to use  with ``dockerize_pip``. Must
                provide exactly one of ``docker_file``, ``docker_image``, or
                ``runtime``.

            **dockerize_pip (Optional[Union[str, bool]])**
                Whether to use Docker when preparing package dependencies with
                pip. Can be set to True/False or the special string 'non-linux'
                which will only run on non Linux systems. To use this option
                Docker must be installed.

            **exclude (Optional[Union[str, List[str]]])**
                Pattern or list of patterns of files to exclude from the
                payload. If provided, any files that match will be ignored,
                regardless of whether they match an inclusion pattern.

                Commonly ignored files are already excluded by default,
                such as ``.git``, ``.svn``, ``__pycache__``, ``*.pyc``,
                ``.gitignore``, etc.

            **include (Optional[Union[str, List[str]]])**
                Pattern or list of patterns of files to include in the
                payload. If provided, only files that match these
                patterns will be included in the payload.

                Omitting it is equivalent to accepting all files that are
                not otherwise excluded.

            **path (str)**
                Base directory of the Lambda function payload content.
                If it not an absolute path, it will be considered relative
                to the directory containing the CFNgin configuration file
                in use.

                Files in this directory will be added to the payload ZIP,
                according to the include and exclude patterns. If not
                patterns are provided, all files in this directory
                (respecting default exclusions) will be used.

                Files are stored in the archive with path names relative to
                this directory. So, for example, all the files contained
                directly under this directory will be added to the root of
                the ZIP file.

            **pipenv_lock_timeout (Optional[int])**
                Time in seconds to wait while creating lock file with pipenv.

            **pipenv_timeout (Optional[int])**
                Time in seconds to wait while running pipenv.

            **python_path (Optional[str])**
                Absolute path to a python interpreter to use for
                ``pip``/``pipenv`` actions. If not provided, the current
                python interpreter will be used for ``pip`` and ``pipenv``
                will be used from the current ``$PATH``.

            **runtime (Optional[str])**
                Runtime of the AWS Lambda Function being uploaded. Used with
                ``dockerize_pip`` to automatically select the appropriate
                Docker image to use. Must provide exactly one of
                ``docker_file``, ``docker_image``, or ``runtime``.

            **use_pipenv (Optional[bool])**
                Explicitly use Pipfile/Pipfile.lock to prepare package
                dependencies even if a requirements.txt file is found.

    Examples:
        .. Hook configuration.
        .. code-block:: yaml

            pre_build:
              upload_functions:
                path: runway.cfngin.hooks.aws_lambda.upload_lambda_functions
                required: true
                enabled: true
                data_key: lambda
                args:
                  bucket: custom-bucket
                  follow_symlinks: true
                  prefix: cloudformation-custom-resources/
                  payload_acl: authenticated-read
                  functions:
                    MyFunction:
                      path: ./lambda_functions
                      dockerize_pip: non-linux
                      use_pipenv: true
                      runtime: python3.8
                      include:
                        - '*.py'
                        - '*.txt'
                      exclude:
                        - '*.pyc'
                        - test/

        .. Blueprint usage
        .. code-block:: python

            from troposphere.awslambda import Function
            from runway.cfngin.blueprints.base import Blueprint

            class LambdaBlueprint(Blueprint):
                def create_template(self):
                    code = self.context.hook_data['lambda']['MyFunction']

                    self.template.add_resource(
                        Function(
                            'MyFunction',
                            Code=code,
                            Handler='my_function.handler',
                            Role='...',
                            Runtime='python2.7'
                        )
                    )

    """
    # TODO add better handling for misconfiguration (e.g. forgetting function names)
    # TODO support defining dockerize_pip options at the top level of args
    custom_bucket = kwargs.get('bucket')
    if not custom_bucket:
        bucket_name = context.bucket_name
        LOGGER.info("lambda: using default bucket from CFNgin: %s",
                    bucket_name)
    else:
        bucket_name = custom_bucket
        LOGGER.info("lambda: using custom bucket: %s", bucket_name)

    custom_bucket_region = kwargs.get("bucket_region")
    if not custom_bucket and custom_bucket_region:
        raise ValueError("Cannot specify `bucket_region` without specifying "
                         "`bucket`.")

    bucket_region = select_bucket_region(
        custom_bucket,
        custom_bucket_region,
        context.config.cfngin_bucket_region,
        provider.region
    )

    # Check if we should walk / follow symlinks
    follow_symlinks = kwargs.get('follow_symlinks', False)
    if not isinstance(follow_symlinks, bool):
        raise ValueError('follow_symlinks option must be a boolean')

    # Check for S3 object acl. Valid values from:
    # https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
    payload_acl = kwargs.get('payload_acl', 'private')

    # Always use the global client for s3
    session = get_session(bucket_region)
    s3_client = session.client('s3')

    ensure_s3_bucket(s3_client, bucket_name, bucket_region)

    prefix = kwargs.get('prefix', '')

    results = {}
    for name, options in kwargs['functions'].items():
        sys_path = (os.path.dirname(context.config_path)
                    if os.path.isfile(context.config_path)
                    else context.config_path)
        results[name] = _upload_function(s3_client, bucket_name, prefix, name,
                                         options, follow_symlinks, payload_acl,
                                         sys_path)

    return results



"""Tests for runway.cfngin.stack."""
import unittest

from mock import MagicMock

from runway.cfngin.config import Config
from runway.cfngin.context import Context
from runway.cfngin.lookups import register_lookup_handler
from runway.cfngin.stack import Stack

from .factories import generate_definition


class TestStack(unittest.TestCase):
    """Tests for runway.cfngin.stack.Stack."""

    def setUp(self):
        """Run before tests."""
        self.sd = {"name": "test"}  # pylint: disable=invalid-name
        self.config = Config({"namespace": "namespace"})
        self.context = Context(config=self.config)
        self.stack = Stack(
            definition=generate_definition("vpc", 1),
            context=self.context,
        )
        register_lookup_handler("noop", lambda **kwargs: "test")

    def test_stack_requires(self):
        """Test stack requires."""
        definition = generate_definition(
            base_name="vpc",
            stack_id=1,
            variables={
                "Var1": "${noop fakeStack3::FakeOutput}",
                "Var2": (
                    "some.template.value:${output fakeStack2::FakeOutput}:"
                    "${output fakeStack::FakeOutput}"
                ),
                "Var3": "${output fakeStack::FakeOutput},"
                        "${output fakeStack2::FakeOutput}",
            },
            requires=["fakeStack"],
        )
        stack = Stack(definition=definition, context=self.context)
        self.assertEqual(len(stack.requires), 2)
        self.assertIn(
            "fakeStack",
            stack.requires,
        )
        self.assertIn(
            "fakeStack2",
            stack.requires,
        )

    def test_stack_requires_circular_ref(self):
        """Test stack requires circular ref."""
        definition = generate_definition(
            base_name="vpc",
            stack_id=1,
            variables={
                "Var1": "${output vpc.1::FakeOutput}",
            },
        )
        stack = Stack(definition=definition, context=self.context)
        with self.assertRaises(ValueError):
            stack.requires  # pylint: disable=pointless-statement

    def test_stack_cfn_parameters(self):
        """Test stack cfn parameters."""
        definition = generate_definition(
            base_name="vpc",
            stack_id=1,
            variables={
                "Param1": "${output fakeStack::FakeOutput}",
            },
        )
        stack = Stack(definition=definition, context=self.context)
        # pylint: disable=protected-access
        stack._blueprint = MagicMock()
        stack._blueprint.get_parameter_values.return_value = {
            "Param2": "Some Resolved Value",
        }
        self.assertEqual(len(stack.parameter_values), 1)
        param = stack.parameter_values["Param2"]
        self.assertEqual(param, "Some Resolved Value")

    def test_stack_tags_default(self):
        """Test stack tags default."""
        self.config.tags = {"environment": "prod"}
        definition = generate_definition(
            base_name="vpc",
            stack_id=1
        )
        stack = Stack(definition=definition, context=self.context)
        self.assertEqual(stack.tags, {"environment": "prod"})

    def test_stack_tags_override(self):
        """Test stack tags override."""
        self.config.tags = {"environment": "prod"}
        definition = generate_definition(
            base_name="vpc",
            stack_id=1,
            tags={"environment": "stage"}
        )
        stack = Stack(definition=definition, context=self.context)
        self.assertEqual(stack.tags, {"environment": "stage"})

    def test_stack_tags_extra(self):
        """Test stack tags extra."""
        self.config.tags = {"environment": "prod"}
        definition = generate_definition(
            base_name="vpc",
            stack_id=1,
            tags={"app": "graph"}
        )
        stack = Stack(definition=definition, context=self.context)
        self.assertEqual(stack.tags, {"environment": "prod", "app": "graph"})


if __name__ == '__main__':
    unittest.main()



"""CFNgin hook for syncing static website to S3 bucket."""
# TODO move to runway.cfngin.hooks on next major release
import logging
import time
from operator import itemgetter

from ...cfngin.lookups.handlers.output import OutputLookup
from ...cfngin.session_cache import get_session
from ...commands.runway.run_aws import aws_cli

LOGGER = logging.getLogger(__name__)


def get_archives_to_prune(archives, hook_data):
    """Return list of keys to delete.

    Keyword Args:
        archives (Dict): The full list of file archives
        hook_data (Dict): CFNgin hook data

    """
    files_to_skip = []

    for i in ['current_archive_filename', 'old_archive_filename']:
        if hook_data.get(i):
            files_to_skip.append(hook_data[i])

    archives.sort(key=itemgetter('LastModified'), reverse=False)  # sort from oldest to newest

    # Drop all but last 15 files
    return [i['Key'] for i in archives[:-15] if i['Key'] not in files_to_skip]


def sync(context, provider, **kwargs):
    """Sync static website to S3 bucket.

    Keyword Args:

        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance.

    """
    session = get_session(provider.region)
    bucket_name = OutputLookup.handle(kwargs.get('bucket_output_lookup'),
                                      provider=provider,
                                      context=context)

    if context.hook_data['staticsite']['deploy_is_current']:
        LOGGER.info('staticsite: skipping upload; latest version already '
                    'deployed')
        if kwargs.get('cf_disabled', '') == 'true':
            display_static_website_url(kwargs.get('website_url'), provider, context)
    else:
        # Using the awscli for s3 syncing is incredibly suboptimal, but on
        # balance it's probably the most stable/efficient option for syncing
        # the files until https://github.com/boto/boto3/issues/358 is resolved
        aws_cli(['s3',
                 'sync',
                 context.hook_data['staticsite']['app_directory'],
                 "s3://%s/" % bucket_name,
                 '--delete'])

        if kwargs.get('cf_disabled', False):
            display_static_website_url(kwargs.get('website_url'), provider, context)
        else:
            distribution = get_distribution_data(context, provider, **kwargs)
            invalidate_distribution(session, **distribution)

        LOGGER.info("staticsite: sync " "complete")

        update_ssm_hash(context, session)

    prune_archives(context, session)
    return True


def display_static_website_url(website_url_handle, provider, context):
    """Based on the url handle display the static website url.

    Keyword Args:
        website_url_handle (str): the Output handle for the website url
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance.
        context (:class:`runway.cfngin.context.Context`): context instance

    """
    bucket_url = OutputLookup.handle(website_url_handle,
                                     provider=provider,
                                     context=context)
    LOGGER.info("STATIC WEBSITE URL: %s", bucket_url)


def update_ssm_hash(context, session):
    """Update the SSM hash with the new tracking data.

    Keyword Args:
        context (:class:`runway.cfngin.context.Context`): context instance
        session (:class:`runway.cfngin.session.Session`): CFNgin session

    """
    if not context.hook_data['staticsite'].get('hash_tracking_disabled'):
        LOGGER.info("staticsite: updating environment SSM parameter %s "
                    "with hash %s",
                    context.hook_data['staticsite']['hash_tracking_parameter'],  # noqa
                    context.hook_data['staticsite']['hash'])
        ssm_client = session.client('ssm')
        ssm_client.put_parameter(
            Name=context.hook_data['staticsite']['hash_tracking_parameter'],  # noqa
            Description='Hash of currently deployed static website source',
            Value=context.hook_data['staticsite']['hash'],
            Type='String',
            Overwrite=True
        )
    return True


def get_distribution_data(context, provider, **kwargs):
    """Retrieve information about the distribution.

    Keyword Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        provider (:class:`runway.cfngin.providers.base.BaseProvider`):
            The provider instance

    """
    LOGGER.info("Retrieved distribution data")
    return {
        'identifier': OutputLookup.handle(
            kwargs.get('distributionid_output_lookup'),
            provider=provider,
            context=context
        ),
        'domain': OutputLookup.handle(
            kwargs.get('distributiondomain_output_lookup'),
            provider=provider,
            context=context
        ),
        'path': kwargs.get('distribution_path', '/*')
    }


def invalidate_distribution(session, identifier='', path='', domain='', **_):
    """Invalidate the current distribution.

    Keyword Args:
        session (Session): The current CFNgin session
        identifier (string): The distribution id
        path (string): The distribution path
        domain (string): The distribution domain

    """
    LOGGER.info("staticsite: Invalidating CF distribution")
    cf_client = session.client('cloudfront')
    cf_client.create_invalidation(
        DistributionId=identifier,
        InvalidationBatch={
            'Paths': {
                'Quantity': 1,
                'Items': [path]},
            'CallerReference': str(time.time())}
    )

    LOGGER.info("staticsite: CF invalidation of %s (domain %s) " "complete", identifier, domain)
    return True


def prune_archives(context, session):
    """Prune the archives from the bucket.

    Keyword Args:
        context (:class:`runway.cfngin.context.Context`): The context
            instance.
        session (:class:`runway.cfngin.session.Session`): The CFNgin
            session.

    """
    LOGGER.info("staticsite: cleaning up old site archives...")
    archives = []
    s3_client = session.client('s3')
    list_objects_v2_paginator = s3_client.get_paginator('list_objects_v2')
    response_iterator = list_objects_v2_paginator.paginate(
        Bucket=context.hook_data['staticsite']['artifact_bucket_name'],
        Prefix=context.hook_data['staticsite']['artifact_key_prefix']
    )

    for page in response_iterator:
        archives.extend(page.get('Contents', []))
    archives_to_prune = get_archives_to_prune(
        archives,
        context.hook_data['staticsite']
    )

    # Iterate in chunks of 1000 to match delete_objects limit
    for objects in [archives_to_prune[i:i + 1000]
                    for i in range(0, len(archives_to_prune), 1000)]:
        s3_client.delete_objects(
            Bucket=context.hook_data['staticsite']['artifact_bucket_name'],
            Delete={'Objects': [{'Key': i} for i in objects]}
        )
    return True



"""Test tfenv."""
import os
from subprocess import check_output

import boto3

from integration_tests.test_commands.test_commands import Commands


class TestRunTFEnv(Commands):
    """Tests run-tfenv subcommand."""

    TEST_NAME = __name__

    def init(self):
        """Initialize test."""
        pass  # pylint: disable=unnecessary-pass

    def get_path(self):
        """Gets the test path."""
        return os.path.join(
            os.path.dirname(os.path.dirname(__file__)),
            'fixtures',
            'tfenv')

    def run(self):
        """Run tests."""
        self.set_env_var('AWS_DEFAULT_REGION', 'us-east-1')
        # init
        check_output(['runway',
                      'tfenv',
                      'run',
                      'init',
                      self.get_path()]).decode()

        # apply
        check_output(['runway',
                      'tfenv',
                      'run',
                      '--',
                      'apply',
                      '-auto-approve',
                      self.get_path()]).decode()

        # output
        key = check_output(['runway',
                            'tfenv',
                            'run',
                            '--',
                            'output',
                            'key']).decode()
        self.logger.info('Key: %s', key)

        # ssm parameter
        client = boto3.client('ssm',
                              region_name=self.environment['AWS_DEFAULT_REGION'])
        parameter = client.get_parameter(Name=key)
        value = parameter['Parameter']['Value']

        assert value == 'bar'

    def teardown(self):
        """Teardown any created resources."""
        check_output(['runway',
                      'tfenv',
                      'run',
                      '--',
                      'destroy',
                      '-auto-approve',
                      self.get_path()]).decode()



"""CFNgin config."""
import copy
import logging
import sys
import warnings
from io import StringIO
from string import Template

import yaml
from schematics import Model
from schematics.exceptions import BaseError as SchematicsError
from schematics.exceptions import UndefinedValueError, ValidationError
from schematics.types import (BaseType, BooleanType, DictType, ListType,
                              ModelType, StringType)
from six import text_type

from .. import exceptions
from ..lookups import register_lookup_handler
from ..util import SourceProcessor, merge_map, yaml_to_ordered_dict
from .translators import *  # noqa pylint: disable=wildcard-import

LOGGER = logging.getLogger(__name__)


def render_parse_load(raw_config, environment=None, validate=True):
    """Encapsulate the render -> parse -> validate -> load process.

    Args:
        raw_config (str): The raw CFNgin configuration string.
        environment (Optional[Dict[str, Any]]): Any environment values that
            should be passed to the config.
        validate (bool): If provided, the config is validated before being
            loaded.

    Returns:
        :class:`Config`: The parsed CFNgin config.

    """
    pre_rendered = render(raw_config, environment)

    rendered = process_remote_sources(pre_rendered, environment)

    config = parse(rendered)

    # For backwards compatibility, if the config doesn't specify a namespace,
    # we fall back to fetching it from the environment, if provided.
    if config.namespace is None:
        namespace = environment.get("namespace")
        if namespace:
            LOGGER.warning("DEPRECATION WARNING: specifying namespace in the "
                           "environment is deprecated. See "
                           "https://docs.onica.com/projects/runway/en/"
                           "release/cfngin/config.html#namespace "
                           "for more info.")
            config.namespace = namespace

    if validate:
        config.validate()

    return load(config)


def render(raw_config, environment=None):
    """Render a config, using it as a template with the environment.

    Args:
        raw_config (str): The raw CFNgin configuration string.
        environment (Optional[Dict[str, Any]]): Any environment values that
            should be passed to the config.

    Returns:
        str: The CFNgin configuration populated with any values passed from
            the environment.

    """
    template = Template(raw_config)
    buff = StringIO()
    if not environment:
        environment = {}
    try:
        substituted = template.substitute(environment)
    except KeyError as err:
        raise exceptions.MissingEnvironment(err.args[0])
    except AttributeError as err:
        # gets the missing attr from the error message to result in the same
        # message as a KeyError
        raise exceptions.MissingEnvironment(err.args[0].split("'")[3])
    except ValueError:
        # Support "invalid" placeholders for lookup placeholders.
        substituted = template.safe_substitute(environment)

    if not isinstance(substituted, text_type):
        substituted = substituted.decode('utf-8')

    buff.write(substituted)
    buff.seek(0)
    return buff.read()


def parse(raw_config):
    """Parse a raw yaml formatted CFNgin config.

    Args:
        raw_config (str): The raw CFNgin configuration string in yaml format.

    Returns:
        :class:`Config`: The parsed CFNgin config.

    """
    # Convert any applicable dictionaries back into lists
    # This is necessary due to the move from lists for these top level config
    # values to either lists or OrderedDicts.
    # Eventually we should probably just make them OrderedDicts only.
    config_dict = yaml_to_ordered_dict(raw_config)
    if config_dict:
        for top_level_key in ['stacks', 'pre_build', 'post_build',
                              'pre_destroy', 'post_destroy']:
            top_level_value = config_dict.get(top_level_key)
            if isinstance(top_level_value, dict):
                tmp_list = []
                for key, value in top_level_value.items():
                    tmp_dict = copy.deepcopy(value)
                    if top_level_key == 'stacks':
                        tmp_dict['name'] = key
                    tmp_list.append(tmp_dict)
                config_dict[top_level_key] = tmp_list

    # Top-level excess keys are removed by Config._convert, so enabling strict
    # mode is fine here.
    try:
        return Config(config_dict, strict=True)
    except SchematicsError as err:
        raise exceptions.InvalidConfig(err.errors)


def load(config):
    """Load a CFNgin configuration by modifying syspath, loading lookups, etc.

    Args:
        config (:class:`Config`): The CFNgin config to load.

    Returns:
        :class:`Config`: The CFNgin config provided above.

    """
    if config.sys_path:
        LOGGER.debug("Appending %s to sys.path.", config.sys_path)
        sys.path.append(config.sys_path)
        LOGGER.debug("sys.path is now %s", sys.path)
    if config.lookups:
        for key, handler in config.lookups.items():
            register_lookup_handler(key, handler)

    return config


def dump(config):
    """Dump a CFNgin Config object as yaml.

    Args:
        config (:class:`Config`): The CFNgin Config object.

    Returns:
        str: The yaml formatted CFNgin Config.

    """
    return yaml.safe_dump(
        config.to_primitive(),
        default_flow_style=False,
        encoding='utf-8',
        allow_unicode=True)


def process_remote_sources(raw_config, environment=None):
    """Stage remote package sources and merge in remote configs.

    Args:
        raw_config (str): The raw CFNgin configuration string.
        environment (Optional[Dict, Any]): Any environment values that should
            be passed to the config.

    Returns:
        str: The raw CFNgin configuration string.

    """
    config = yaml.safe_load(raw_config)
    if config and config.get('package_sources'):
        processor = SourceProcessor(
            sources=config['package_sources'],
            cfngin_cache_dir=config.get('cfngin_cache_dir',
                                        config.get('stacker_cache_dir'))
        )
        processor.get_package_sources()
        if processor.configs_to_merge:
            for i in processor.configs_to_merge:
                LOGGER.debug("Merging in remote config \"%s\"", i)
                remote_config = yaml.safe_load(open(i))
                config = merge_map(remote_config, config)
            # Call the render again as the package_sources may have merged in
            # additional environment lookups
            if not environment:
                environment = {}
            return render(str(config), environment)

    return raw_config


class AnyType(BaseType):
    """Any type."""


class LocalPackageSource(Model):
    """Local package source model.

    Package source located on a local disk.

    Attributes:
        configs (ListType): List of CFNgin config paths to execute.
        paths (ListType): List of paths to append to ``sys.path``.
        source (StringType): Source.

    """

    configs = ListType(StringType, serialize_when_none=False)
    paths = ListType(StringType, serialize_when_none=False)
    source = StringType(required=True)


class GitPackageSource(Model):
    """Git package source model.

    Package source located in a git repo.

    Attributes:
        branch (StringType): Branch name.
        commit (StringType): Commit hash.
        configs (ListType): List of CFNgin config paths to execute.
        paths (ListType): List of paths to append to ``sys.path``.
        tag (StringType): Git tag.
        uri (StringType): Remote git repo URI.

    """

    branch = StringType(serialize_when_none=False)
    commit = StringType(serialize_when_none=False)
    configs = ListType(StringType, serialize_when_none=False)
    paths = ListType(StringType, serialize_when_none=False)
    tag = StringType(serialize_when_none=False)
    uri = StringType(required=True)


class S3PackageSource(Model):
    """S3 package source model.

    Package source located in AWS S3.

    Attributes:
        bucket (StringType): AWS S3 bucket name.
        configs (ListType): List of CFNgin config paths to execute.
        key (StringType): Object key. The object should be a zip file.
        paths (ListType): List of paths to append to ``sys.path``.
        requester_pays (BooleanType): AWS S3 requester pays option.
        use_latest (BooleanType): Use the latest version of the object.

    """

    bucket = StringType(required=True)
    configs = ListType(StringType, serialize_when_none=False)
    key = StringType(required=True)
    paths = ListType(StringType, serialize_when_none=False)
    requester_pays = BooleanType(serialize_when_none=False)
    use_latest = BooleanType(serialize_when_none=False)


class PackageSources(Model):
    """Package sources model.

    Attributes:
        git (GitPackageSource): Package source located in a git repo.
        local (LocalPackageSource): Package source located on a local disk.
        s3 (S3PackageSource): Package source located in AWS S3.

    """

    git = ListType(ModelType(GitPackageSource))
    local = ListType(ModelType(LocalPackageSource))
    s3 = ListType(ModelType(S3PackageSource))


class Hook(Model):
    """Hook module.

    Attributes:
        args (DictType)
        data_key (StringType)
        enabled (BooleanType)
        path (StringType)
        required (BooleanType)

    """

    args = DictType(AnyType)
    data_key = StringType(serialize_when_none=False)
    enabled = BooleanType(default=True)
    path = StringType(required=True)
    required = BooleanType(default=True)


class Target(Model):
    """Target model.

    Attributes:
        name (StringType)
        required_by (ListType)
        requires (ListType)

    """

    name = StringType(required=True)
    required_by = ListType(StringType, serialize_when_none=False)
    requires = ListType(StringType, serialize_when_none=False)


class Stack(Model):
    """Stack model.

    Attributes:
        class_path (StringType)
        description (StringType)
        enabled (BooleanType)
        in_progress_behavior (StringType)
        locked (BooleanType)
        name (StringType)
        parameters (DictType)
        profile (StringType)
        protected (BooleanType)
        region (StringType)
        required_by (ListType)
        requires (ListType)
        stack_name (StringType)
        stack_policy_path (StringType)
        tags (DictType)
        template_path (StringType)
        termination_protection (BooleanType)
        variables (DictType)

    """

    class_path = StringType(serialize_when_none=False)
    description = StringType(serialize_when_none=False)
    enabled = BooleanType(default=True)
    in_progress_behavior = StringType(serialize_when_none=False)
    locked = BooleanType(default=False)
    name = StringType(required=True)
    parameters = DictType(AnyType, serialize_when_none=False)
    profile = StringType(serialize_when_none=False)
    protected = BooleanType(default=False)
    region = StringType(serialize_when_none=False)
    required_by = ListType(StringType, serialize_when_none=False)
    requires = ListType(StringType, serialize_when_none=False)
    stack_name = StringType(serialize_when_none=False)
    stack_policy_path = StringType(serialize_when_none=False)
    tags = DictType(StringType, serialize_when_none=False)
    template_path = StringType(serialize_when_none=False)
    termination_protection = BooleanType(default=False)
    variables = DictType(AnyType, serialize_when_none=False)

    def validate_class_path(self, data, value):
        """Validate class pass."""
        if value and data["template_path"]:
            raise ValidationError(
                "template_path cannot be present when "
                "class_path is provided.")
        self.validate_stack_source(data)

    def validate_template_path(self, data, value):
        """Validate template path."""
        if value and data["class_path"]:
            raise ValidationError(
                "class_path cannot be present when "
                "template_path is provided.")
        self.validate_stack_source(data)

    @staticmethod
    def validate_stack_source(data):
        """Validate stack source."""
        # Locked stacks don't actually need a template, since they're
        # read-only.
        if data["locked"]:
            return

        if not (data["class_path"] or data["template_path"]):
            raise ValidationError(
                "class_path or template_path is required.")

    def validate_parameters(self, data, value):  # pylint: disable=no-self-use
        """Validate parameters."""
        if value:
            stack_name = data['name']
            raise ValidationError(
                "DEPRECATION: Stack definition %s contains "
                "deprecated 'parameters', rather than 'variables'. You are"
                " required to update your config. See "
                "https://docs.onica.com/projects/runway/en/release/cfngin/"
                "config.html#variables for additional information."
                % stack_name)
        return value


class Config(Model):
    """This is the Python representation of a CFNgin config file.

    This is used internally by CFNgin to parse and validate a yaml formatted
    CFNgin configuration file, but can also be used in scripts to generate a
    CFNgin config file before handing it off to CFNgin to build/destroy.

    Example::

        from runway.cfngin.config import dump, Config, Stack

        vpc = Stack({
            "name": "vpc",
            "class_path": "blueprints.VPC"})

        config = Config()
        config.namespace = "prod"
        config.stacks = [vpc]

        print dump(config)

    Attributes:
        cfngin_bucket (StringType): Bucket to use for CFNgin resources (e.g.
            CloudFormation templates). May be an empty string.
        cfngin_bucket_region (StringType): Explicit region to use for
            ``cfngin_bucket``.
        cfngin_cache_dir (StringType): Local directory to use for caching.
        log_formats (DictType): Custom formatting for log messages.
        lookups (DictType): Register custom lookups.
        mappings (DictType): Mappings that will be added to all stacks.
        namespace (StringType): Namespace to prepend to everything.
        namespace_delimiter (StringType): Character used to separate
            ``namespace`` and anything it prepends.
        package_sources (ModelType): Remote source locations.
        persistent_graph_key (str): S3 object key were the persistent graph
            is stored.
        post_build (ListType): Hooks to run after a build action.
        post_destroy (ListType): Hooks to run after a destroy action.
        pre_build (ListType): Hooks to run before a build action.
        pre_destroy (ListType): Hooks to run before a destroy action.
        service_role (StringType): IAM role for CloudFormation to use.
        stacker_bucket (StringType): [DEPRECATED] Replaced by
            ``cfngin_bucket``, support will be retained until the release
            of version 2.0.0 at the earliest.
        stacker_bucket_region (StringType): [DEPRECATED] Replaced by
            ``cfngin_bucket_region``, support will be retained until the
            release of version 2.0.0 at the earliest.
        stacker_cache_dir (StringType): [DEPRECATED] Replaced by
            ``cfngin_cache_dir``, support will be retained until the release
            of version 2.0.0 at the earliest.
        stacks (ListType): Stacks to be processed.
        sys_path (StringType): Relative or absolute path to use as the work
            directory.
        tags (DictType): Tags to apply to all resources.
        targets (ListType): Stag grouping.
        template_indent (StringType): Spaces to use per-indent level when
            outputing a template to json.

    """

    cfngin_bucket = StringType(serialize_when_none=False)
    cfngin_bucket_region = StringType(serialize_when_none=False)
    cfngin_cache_dir = StringType(serialize_when_none=False)
    log_formats = DictType(StringType, serialize_when_none=False)
    lookups = DictType(StringType, serialize_when_none=False)
    mappings = DictType(
        DictType(DictType(StringType)), serialize_when_none=False)
    namespace = StringType(required=True)
    namespace_delimiter = StringType(serialize_when_none=False)
    package_sources = ModelType(PackageSources, serialize_when_none=False)
    persistent_graph_key = StringType(serialize_when_none=False)
    post_build = ListType(ModelType(Hook), serialize_when_none=False)
    post_destroy = ListType(ModelType(Hook), serialize_when_none=False)
    pre_build = ListType(ModelType(Hook), serialize_when_none=False)
    pre_destroy = ListType(ModelType(Hook), serialize_when_none=False)
    service_role = StringType(serialize_when_none=False)
    stacker_bucket = StringType(serialize_when_none=False)
    stacker_bucket_region = StringType(serialize_when_none=False)
    stacker_cache_dir = StringType(serialize_when_none=False)
    stacks = ListType(
        ModelType(Stack), default=[])
    sys_path = StringType(serialize_when_none=False)
    tags = DictType(StringType, serialize_when_none=False)
    targets = ListType(
        ModelType(Target), serialize_when_none=False)
    template_indent = StringType(serialize_when_none=False)

    def __init__(self, raw_data=None, trusted_data=None,
                 deserialize_mapping=None, init=True, partial=True,
                 strict=True, validate=False, app_data=None, lazy=False,
                 **kwargs):
        """Extend functionality of the parent class.

        Manipulation here allows us to _clone_ the values of legacy stacker
        field into their new names. Doing so we can retain support for Stacker
        configs and CFNgin configs.

        """
        if raw_data:  # this can be empty when running unittests
            for field_suffix in ['bucket', 'bucket_region', 'cache_dir']:
                cfngin_field = 'cfngin_' + field_suffix
                stacker_field = 'stacker_' + field_suffix
                # explicitly check for an empty string since it has specific logic.
                # cfngin fields with a value take precedence.
                if (
                        not (raw_data.get(cfngin_field) or
                             raw_data.get(cfngin_field) == '') and
                        (raw_data.get(stacker_field) or
                         raw_data.get(stacker_field) == '')
                ):
                    raw_data[cfngin_field] = raw_data[stacker_field]
        super(Config, self).__init__(raw_data=raw_data,
                                     trusted_data=trusted_data,
                                     deserialize_mapping=deserialize_mapping,
                                     init=init,
                                     partial=partial,
                                     strict=strict,
                                     validate=validate,
                                     app_data=app_data,
                                     lazy=lazy,
                                     **kwargs)

    def _remove_excess_keys(self, data):
        excess_keys = set(data.keys())
        # attribute is defined in __new__ of base class
        excess_keys -= self._schema.valid_input_keys  # pylint: disable=no-member
        if not excess_keys:
            return data

        LOGGER.debug('Removing excess keys from config input: %s',
                     excess_keys)
        clean_data = data.copy()
        for key in excess_keys:
            del clean_data[key]

        return clean_data

    def _convert(self, raw_data=None, context=None, **kwargs):
        if raw_data is not None:
            # Remove excess top-level keys, since we want to allow them to be
            # used for custom user variables to be reference later. This is
            # preferable to just disabling strict mode, as we can still
            # disallow excess keys in the inner models.
            raw_data = self._remove_excess_keys(raw_data)

        return super(Config, self)._convert(raw_data=raw_data, context=context,
                                            **kwargs)

    def validate(self, partial=False, convert=True, app_data=None, **kwargs):
        """Validate the state of the model.

        If the data is invalid, raises a ``DataError`` with error messages.

        Args:
            partial (bool): Allow partial data to validate. Essentially drops
                the ``required=True`` settings from field definitions.
            convert (bool): Controls whether to perform import conversion
                before validating. Can be turned off to skip an unnecessary
                conversion step if all values are known to have the right
                datatypes (e.g., when validating immediately after the initial
                import).

        Raises:
            UndefinedValueError
            SchematicsError

        """
        try:
            return super(Config, self).validate(partial, convert, app_data,
                                                **kwargs)
        except UndefinedValueError as err:
            raise exceptions.InvalidConfig([str(err)])
        except SchematicsError as err:
            raise exceptions.InvalidConfig(err.errors)

    def validate_stacker_bucket(self, _data, value):  # pylint: disable=no-self-use
        """Validate stack_bucket is not used.

        If in use, show deprecation warning.

        """
        msg = ('Use of "stacker_bucket" has been deprecated and will be '
               'removed after the next major release. Please use '
               '"cfngin_bucket".')
        if value or value == '':
            warnings.warn(msg, DeprecationWarning)
            LOGGER.warning(msg)

    def validate_stacker_bucket_region(self, _data, value):  # pylint: disable=no-self-use
        """Validate stacker_bucket_regio is not used.

        If in use, show deprecation warning.

        """
        msg = ('Use of "stacker_bucket_region" has been deprecated and will '
               'be removed after the next major release. Please use '
               '"cfngin_bucket_region".')
        if value:
            warnings.warn(msg, DeprecationWarning)
            LOGGER.warning(msg)

    def validate_stacker_cache_dir(self, _data, value):  # pylint: disable=no-self-use
        """Validate stacker_cache_dir is not used.

        If in use, show deprecation warning.

        """
        msg = ('Use of "stacker_cache_dir" has been deprecated and will '
               'be removed after the next major release. Please use '
               '"cfngin_cache_dir".')
        if value:
            warnings.warn(msg, DeprecationWarning)
            LOGGER.warning(msg)

    def validate_stacks(self, _data, value):  # pylint: disable=no-self-use
        """Validate stacks."""
        if value:
            stack_names = [stack.name for stack in value]
            if len(set(stack_names)) != len(stack_names):
                # only loop / enumerate if there is an issue.
                for i, stack_name in enumerate(stack_names):
                    if stack_names.count(stack_name) != 1:
                        raise ValidationError(
                            "Duplicate stack %s found at index %d."
                            % (stack_name, i))



""".. Base class for lookup handlers.

.. _lookup arguments:

Lookup Arguments
^^^^^^^^^^^^^^^^

Arguments can be passed to Lookups to effect how they function.

To provide arguments to a Lookup, use a double-colon (``::``) after the
query. Each argument is then defined as a **key** and **value** seperated with
equals (``=``) and the arguments theselves are seperated with a comma (``,``).
The arguments can have an optional space after the comma and before the next
key to make them easier to read but this is not required. The value of all
arguments are read as strings.

.. rubric:: Example
.. code-block:: yaml

    ${var my_query::default=true, transform=bool}
    ${env MY_QUERY::default=1,transform=bool}

Each Lookup may have their own, specific arguments that it uses to modify its
functionality or the value it returns. There is also a common set of arguments
that all Lookups accept.

.. _Common Lookup Arguments:

Common Lookup Arguments
~~~~~~~~~~~~~~~~~~~~~~~

**default (Any)**
    If the Lookup is unable to find a value for the provided query, this
    value will be returned instead of raising an exception.

**get (Optional[str])**
    Can be used on a dictionary type object to retrieve a specific piece of
    data. This is executed after the optional ``load`` step.

**indent (Optional[int])**
    Number of spaces to use per indent level when transforming a dictionary
    type object to a string.

**load (Optional[str])**
    Load the data to be processed by a Lookup using a specific parser. This is
    the first action taking on the data after it has been retrieved from its
    source. The data must be in a format that is supported by the parser
    in order for it to be used.

    **json**
        Loads a JSON seralizable string into a dictionary like object.
    **troposphere**
        Loads the ``properties`` of a subclass of ``troposphere.BaseAWSObject``
        into a dictinary.
    **yaml**
        Loads a YAML seralizable string into a dictionary like object.

**region (Optional[str])**
    AWS region used when creating a ``boto3.Session`` to retrieve data.
    If not provided, the region currently being processed will be used.
    This can be specified to always get data from one region regardless of
    region is being deployed to.

**transform (Optional[str])**
    Transform the data that will be returned by a Lookup into a different
    data type. This is the last action taking on the data before it is
    returned. Supports the following:

    **str**
        Converts any value to a string. The original data type determines the
        end result.

        ``list``, ``set``, and ``tuple`` will become a comma delimited list

        ``dict`` and anything else will become an escaped JSON string.
    **bool**
        Converts a string or boolean value into a boolean.

.. rubric:: Example
.. code-block:: yaml

  deployments:
    - environments:
        some_variable: ${var some_value::default=my_value}
        comma_list: ${var my_list::default=undefined, transform=str}

"""
import json
import logging
# python2 supported pylint is unable to load this when in a venv
from distutils.util import strtobool  # pylint: disable=E
from typing import (TYPE_CHECKING, Any, Dict,  # noqa: F401 pylint: disable=W
                    Optional, Tuple, Union)

import yaml
from six import string_types
from troposphere import BaseAWSObject

from runway.cfngin.util import read_value_from_path
from runway.util import MutableMap

# python2 supported pylint sees this is cyclic even though its only for type checking
# pylint: disable=cyclic-import
if TYPE_CHECKING:
    from ...context import Context  # noqa: F401 pylint: disable=unused-import

LOGGER = logging.getLogger(__name__)


class LookupHandler(object):
    """Base class for lookup handlers."""

    @classmethod
    def dependencies(cls, _lookup_data):
        """Calculate any dependencies required to perform this lookup.

        Note that lookup_data may not be (completely) resolved at this time.

        Args:
            lookup_data (VariableValue): Parameter(s) given to this lookup.

        Returns:
            Set

        """
        return set()

    @classmethod
    def format_results(cls,
                       value,  # type: Any
                       get=None,  # type: Optional[str]
                       load=None,  # type: Optional[str]
                       transform=None,  # type: Optional[str]
                       **kwargs  # type: Any
                       ):  # python2 only # pylint: disable=bad-continuation
        # type: (...) -> Any
        """Format results to be returned by a lookup.

        Args:
            value (Any): Data collected by the Lookup.
            get (Optional[str]): Nested value to get from a dictionary like
                object.
            load (Optional[str]): Parser to use to parse a formatted string
                before the ``get`` and ``transform`` method.
            transform (Optional[str]): Convert the final value to a different
                data type before returning it.

        Raises:
            TypeError: If ``get`` is provided but the value value is not a
                dictionary like object.

        Runs the following actions in order:

        1. :meth:`~LookupHandler.load` if ``load`` is provided.
        2. :meth:`runway.util.MutableMap.find` or :meth:`dict.get` depending
           on the data type if ``get`` is provided.
        3. :meth:`~LookupHandler.transform` if ``transform`` is provided.

        """
        if load:
            value = cls.load(value, parser=load, **kwargs)
        if get:
            if isinstance(value, MutableMap):
                value = value.find(get)
            elif isinstance(value, dict):
                value = value.get(get)
            else:
                raise TypeError('value must be dict type to use "get"; got '
                                'type "{}"'.format(type(value)))
        if transform:
            return cls.transform(value, to_type=transform, **kwargs)
        if isinstance(value, MutableMap):
            LOGGER.debug('Returning data from MutableMap')
            return value.data
        return value

    @classmethod
    def handle(cls, value, context, **kwargs):
        # type: (str, 'Context', Any) -> Any
        """Perform the lookup.

        Args:
            value: Parameter(s) given to the lookup.
            context: The current context object.
            provider: Optional provider to use when handling the lookup.

        Returns:
            (Any) Looked-up value.

        """
        raise NotImplementedError

    @classmethod
    def parse(cls, value):
        # type: (str) -> Tuple[str, Dict[str, str]]
        """Parse the value passed to a lookup in a standardized way.

        Args:
            value: The raw value passed to a lookup.

        Returns:
            The lookup query and a dict of arguments

        """
        raw_value = read_value_from_path(value)

        colon_split = raw_value.split('::', 1)

        query = colon_split.pop(0)
        args = cls._parse_args(colon_split[0]) if colon_split else {}

        return query, args

    @classmethod
    def _parse_args(cls, args):
        # type: (str) -> Dict[str, str]
        """Convert a string into an args dict.

        Each arg should be seporated by  ``,``. The key and value should
        be seporated by ``=``. Any leading or following spaces are stripped.

        Args:
            args: A string containing arguments to be parsed. (e.g.
                ``'key1=value1, key2=value2'``)

        Returns:
            Dict of parsed args.

        """
        split_args = args.split(',')
        return {key.strip(): value.strip() for key, value in
                [arg.split('=') for arg in split_args]}

    @classmethod
    def load(cls, value, parser=None, **kwargs):
        # type: (str, str, Any) -> Any
        """Load a formatted string or object into a python data type.

        First action taken in :meth:`~LookupHandler.format_results`.
        If a lookup needs to handling loading data to process it before it
        enters :meth:`~LookupHandler.format_results`, is should use
        ``args.pop('load')`` to prevent the data from being loaded twice.

        Args:
            value: What is being loaded.
            parser: Name of the parser to use.

        Returns:
            The loaded value.

        """
        mapping = {
            'json': cls._load_json,
            'troposphere': cls._load_troposphere,
            'yaml': cls._load_yaml
        }

        if not parser:
            return value

        return mapping[parser](value, **kwargs)

    @classmethod
    def _load_json(cls, value, **_):
        # type: (str, Any) -> MutableMap
        """Load a JSON string into a MutableMap.

        Args:
            value: JSON formatted string.

        Returns:
            MutableMap

        """
        if not isinstance(value, str):
            raise TypeError('value of type "%s" must of type "str" to use '
                            'the "load=json" argument.')
        result = json.loads(value)
        if isinstance(result, dict):
            return MutableMap(**result)
        return result

    @classmethod
    def _load_troposphere(cls, value, **_):
        # type: (BaseAWSObject, Any) -> MutableMap
        """Load a Troposphere resource into a MutableMap.

        Args:
            Value (troposphere.BaseAWSObject): Troposphere resource to
                contvert to a MutableMap for parsing.

        Returns:
            MutableMap

        """
        if not isinstance(value, BaseAWSObject):
            raise TypeError('value of type "%s" must of type "troposphere.'
                            'BaseAWSObject" to use the "load=troposphere" '
                            'option.')
        if hasattr(value, 'properties'):
            return MutableMap(**value.properties)
        raise NotImplementedError('"load=troposphere" only supports '
                                  'BaseAWSObject with a "properties" '
                                  'object.')

    @classmethod
    def _load_yaml(cls, value, **_):
        # type: (str, Any) -> MutableMap
        """Load a YAML string into a MutableMap.

        Args:
            value: YAML formatted string.

        Returns:
            MutableMap

        """
        if not isinstance(value, str):
            raise TypeError('value of type "%s" must of type "str" to use '
                            'the "load=yaml" argument.')
        result = yaml.safe_load(value)
        if isinstance(result, dict):
            return MutableMap(**result)
        return result

    @classmethod
    def transform(cls, value, to_type='str', **kwargs):
        # type: (str, str, Any) -> Any
        """Transform the result of a lookup into another datatype.

        Last action taken in :meth:`~LookupHandler.format_results`.
        If a lookup needs to handling transforming the data in a way that
        the base class can't support it should overwrite this method of the
        base class to register different transform methods.

        Args:
            value: What is to be transformed.
            to_type: The type the value will be transformed into.

        Returns:
            The transformed value.

        """
        mapping = {
            'bool': cls._transform_to_bool,
            'str': cls._transform_to_string
        }

        if not to_type:
            return value

        return mapping[to_type](value, **kwargs)  # type: ignore

    @classmethod
    def _transform_to_bool(cls, value, **_):
        # type: (Union[bool, str], Any) -> bool
        """Transform a string into a bool.

        Args:
            value: The value to be transformed into a bool.

        Raises:
            ValueError: The value provided was not a bool or string or
                the string could not be converted to a bool.

        """
        if isinstance(value, bool):
            return value
        if isinstance(value, string_types):
            return bool(strtobool(value))
        raise TypeError('Value must be a string or bool to use transform=bool. '
                        'Got type {}.'.format(type(value)))

    @classmethod
    def _transform_to_string(cls, value, delimiter=None, **kwargs):
        # type: (Any, str, Any) -> str
        """Transform anything into a string.

        If the datatype of ``value`` is a list or similar to a list, ``join()``
        is used to construct the list using a given delimiter or ``,``.

        Args:
            value: The value to be transformed into a string.
            delimited: Used when transforming a list like object into a string
                to join each element together.

        """
        if isinstance(value, (list, set, tuple)):
            return '{}'.format(delimiter or ',').join(value)
        if isinstance(value, MutableMap):
            # convert into a dict with protected attrs removed
            value = value.data
        if isinstance(value, dict):
            # dumped twice for an escaped json dict
            return json.dumps(json.dumps(value,
                                         indent=int(kwargs.get('indent', 0))))
        if isinstance(value, bool):
            return json.dumps(str(value))
        return str(value)



"""Tests for runway.cfngin.util."""
# pylint: disable=unused-argument,invalid-name
import unittest

import boto3
import mock

from runway.cfngin.config import GitPackageSource
from runway.cfngin.util import (Extractor, SourceProcessor, TarExtractor,
                                TarGzipExtractor, ZipExtractor, camel_to_snake,
                                cf_safe_name, get_client_region,
                                get_s3_endpoint, merge_map,
                                parse_cloudformation_template,
                                s3_bucket_location_constraint,
                                yaml_to_ordered_dict)

AWS_REGIONS = ["us-east-1", "cn-north-1", "ap-northeast-1", "eu-west-1",
               "ap-southeast-1", "ap-southeast-2", "us-west-2", "us-gov-west-1",
               "us-west-1", "eu-central-1", "sa-east-1"]


def mock_create_cache_directories(self, **kwargs):
    """Mock create cache directories.

    Don't actually need the directories created in testing

    """
    return 1


class TestUtil(unittest.TestCase):
    """Tests for runway.cfngin.util."""

    def test_cf_safe_name(self):
        """Test cf safe name."""
        tests = (
            ("abc-def", "AbcDef"),
            ("GhI", "GhI"),
            ("jKlm.noP", "JKlmNoP")
        )
        for test in tests:
            self.assertEqual(cf_safe_name(test[0]), test[1])

    def test_camel_to_snake(self):
        """Test camel to snake."""
        tests = (
            ("TestTemplate", "test_template"),
            ("testTemplate", "test_template"),
            ("test_Template", "test__template"),
            ("testtemplate", "testtemplate"),
        )
        for test in tests:
            self.assertEqual(camel_to_snake(test[0]), test[1])

    def test_merge_map(self):
        """Test merge map."""
        tests = [
            # 2 lists of stacks defined
            [{'stacks': [{'stack1': {'variables': {'a': 'b'}}}]},
             {'stacks': [{'stack2': {'variables': {'c': 'd'}}}]},
             {'stacks': [
                 {'stack1': {
                     'variables': {
                         'a': 'b'}}},
                 {'stack2': {
                     'variables': {
                         'c': 'd'}}}]}],
            # A list of stacks combined with a higher precedence dict of stacks
            [{'stacks': [{'stack1': {'variables': {'a': 'b'}}}]},
             {'stacks': {'stack2': {'variables': {'c': 'd'}}}},
             {'stacks': {'stack2': {'variables': {'c': 'd'}}}}],
            # 2 dicts of stacks with non-overlapping variables merged
            [{'stacks': {'stack1': {'variables': {'a': 'b'}}}},
             {'stacks': {'stack1': {'variables': {'c': 'd'}}}},
             {'stacks': {
                 'stack1': {
                     'variables': {
                         'a': 'b',
                         'c': 'd'}}}}],
            # 2 dicts of stacks with overlapping variables merged
            [{'stacks': {'stack1': {'variables': {'a': 'b'}}}},
             {'stacks': {'stack1': {'variables': {'a': 'c'}}}},
             {'stacks': {'stack1': {'variables': {'a': 'c'}}}}],
        ]
        for test in tests:
            self.assertEqual(merge_map(test[0], test[1]), test[2])

    def test_yaml_to_ordered_dict(self):
        """Test yaml to ordered dict."""
        raw_config = """
        pre_build:
          hook2:
            path: foo.bar
          hook1:
            path: foo1.bar1
        """
        config = yaml_to_ordered_dict(raw_config)
        self.assertEqual(list(config['pre_build'].keys())[0], 'hook2')
        self.assertEqual(config['pre_build']['hook2']['path'], 'foo.bar')

    def test_get_client_region(self):
        """Test get client region."""
        regions = ["us-east-1", "us-west-1", "eu-west-1", "sa-east-1"]
        for region in regions:
            client = boto3.client("s3", region_name=region)
            self.assertEqual(get_client_region(client), region)

    def test_get_s3_endpoint(self):
        """Test get s3 endpoint."""
        endpoint_url = "https://example.com"
        client = boto3.client("s3", region_name="us-east-1",
                              endpoint_url=endpoint_url)
        self.assertEqual(get_s3_endpoint(client), endpoint_url)

    def test_s3_bucket_location_constraint(self):
        """Test s3 bucket location constraint."""
        tests = (
            ("us-east-1", ""),
            ("us-west-1", "us-west-1")
        )
        for region, result in tests:
            self.assertEqual(
                s3_bucket_location_constraint(region),
                result
            )

    def test_parse_cloudformation_template(self):
        """Test parse cloudformation template."""
        template = """AWSTemplateFormatVersion: "2010-09-09"
Parameters:
  Param1:
    Type: String
Resources:
  Bucket:
    Type: AWS::S3::Bucket
    Properties:
      BucketName:
        !Join
          - "-"
          - - !Ref "AWS::StackName"
            - !Ref "AWS::Region"
Outputs:
  DummyId:
    Value: dummy-1234"""
        parsed_template = {
            'AWSTemplateFormatVersion': '2010-09-09',
            'Outputs': {'DummyId': {'Value': 'dummy-1234'}},
            'Parameters': {'Param1': {'Type': 'String'}},
            'Resources': {
                'Bucket': {'Type': 'AWS::S3::Bucket',
                           'Properties': {
                               'BucketName': {
                                   u'Fn::Join': [
                                       '-',
                                       [{u'Ref': u'AWS::StackName'},
                                        {u'Ref': u'AWS::Region'}]
                                   ]
                               }
                           }}
            }
        }
        self.assertEqual(
            parse_cloudformation_template(template),
            parsed_template
        )

    def test_extractors(self):
        """Test extractors."""
        self.assertEqual(Extractor('test.zip').archive, 'test.zip')
        self.assertEqual(TarExtractor().extension(), '.tar')
        self.assertEqual(TarGzipExtractor().extension(), '.tar.gz')
        self.assertEqual(ZipExtractor().extension(), '.zip')
        for i in [TarExtractor(), ZipExtractor(), ZipExtractor()]:
            i.set_archive('/tmp/foo')
            self.assertEqual(i.archive.endswith(i.extension()), True)

    def test_SourceProcessor_helpers(self):  # noqa: N802
        """Test SourceProcessor helpers."""
        with mock.patch.object(SourceProcessor,
                               'create_cache_directories',
                               new=mock_create_cache_directories):
            sp = SourceProcessor(sources={})

            self.assertEqual(
                sp.sanitize_git_path('git@github.com:foo/bar.git'),
                'git_github.com_foo_bar'
            )
            self.assertEqual(
                sp.sanitize_uri_path('http://example.com/foo/bar.gz@1'),
                'http___example.com_foo_bar.gz_1'
            )
            self.assertEqual(
                sp.sanitize_git_path('git@github.com:foo/bar.git', 'v1'),
                'git_github.com_foo_bar-v1'
            )

            for i in [GitPackageSource({'branch': 'foo'}), {'branch': 'foo'}]:
                self.assertEqual(
                    sp.determine_git_ls_remote_ref(i),
                    'refs/heads/foo'
                )
            for i in [{'uri': 'git@foo'}, {'tag': 'foo'}, {'commit': '1234'}]:
                self.assertEqual(
                    sp.determine_git_ls_remote_ref(GitPackageSource(i)),
                    'HEAD'
                )
                self.assertEqual(
                    sp.determine_git_ls_remote_ref(i),
                    'HEAD'
                )

            self.assertEqual(
                sp.git_ls_remote('https://github.com/remind101/stacker.git',
                                 'refs/heads/release-1.0'),
                b'857b4834980e582874d70feef77bb064b60762d1'
            )

            bad_configs = [{'uri': 'x',
                            'commit': '1234',
                            'tag': 'v1',
                            'branch': 'x'},
                           {'uri': 'x', 'commit': '1234', 'tag': 'v1'},
                           {'uri': 'x', 'commit': '1234', 'branch': 'x'},
                           {'uri': 'x', 'tag': 'v1', 'branch': 'x'},
                           {'uri': 'x', 'commit': '1234', 'branch': 'x'}]
            for i in bad_configs:
                with self.assertRaises(ImportError):
                    sp.determine_git_ref(GitPackageSource(i))
                with self.assertRaises(ImportError):
                    sp.determine_git_ref(i)

            self.assertEqual(
                sp.determine_git_ref(
                    GitPackageSource({'uri': 'https://github.com/remind101/'
                                             'stacker.git',
                                      'branch': 'release-1.0'})),
                '857b4834980e582874d70feef77bb064b60762d1'
            )
            self.assertEqual(
                sp.determine_git_ref(
                    GitPackageSource({'uri': 'git@foo', 'commit': '1234'})),
                '1234'
            )
            self.assertEqual(
                sp.determine_git_ref({'uri': 'git@foo', 'commit': '1234'}),
                '1234'
            )
            self.assertEqual(
                sp.determine_git_ref(
                    GitPackageSource({'uri': 'git@foo', 'tag': 'v1.0.0'})),
                'v1.0.0'
            )
            self.assertEqual(
                sp.determine_git_ref({'uri': 'git@foo', 'tag': 'v1.0.0'}),
                'v1.0.0'
            )


class MockException1(Exception):
    """Mock exception 1."""


class MockException(Exception):
    """Mock exception 2."""


class TestExceptionRetries(unittest.TestCase):
    """Test exception retries."""

    def setUp(self):
        """Run before tests."""
        self.counter = 0

    def _works_immediately(self, a, b, x=None, y=None):
        """Works immediately."""
        self.counter += 1
        return [a, b, x, y]

    def _works_second_attempt(self, a, b, x=None, y=None):
        """Works second_attempt."""
        self.counter += 1
        if self.counter == 2:
            return [a, b, x, y]
        raise Exception("Broke.")

    def _second_raises_exception2(self, a, b, x=None, y=None):
        """Second raises exception2."""
        self.counter += 1
        if self.counter == 2:
            return [a, b, x, y]
        raise MockException("Broke.")

    def _throws_exception2(self, a, b, x=None, y=None):
        """Throws exception2."""
        self.counter += 1
        raise MockException("Broke.")
